from __future__ import annotations

import pathlib
import pickle

import numpy as np
import pandas as pd
import pytest
from psychopy.data import TrialHandlerExt

from vstt.display import default_display_options
from vstt.experiment import Experiment
from vstt.meta import default_metadata
from vstt.trial import default_trial


def test_experiment_no_results(tmp_path: pathlib.Path) -> None:
    exp = Experiment()
    # default initial values
    assert exp.trial_list == [default_trial()]
    assert exp.metadata == default_metadata()
    assert exp.display_options == default_display_options()
    assert exp.trial_handler_with_results is None
    assert exp.has_unsaved_changes is False
    # get a trial handler for this experiment
    th = exp.create_trialhandler()
    assert th.trialList == [default_trial()]
    assert th.extraInfo["metadata"] == default_metadata()
    assert th.extraInfo["display_options"] == default_display_options()
    assert th.finished is False
    # modify experiment
    exp.trial_list.append(default_trial())
    exp.metadata["author"] = "Joe Smith"
    exp.display_options["to_target_paths"] = False
    assert exp.trial_handler_with_results is None
    # get a new trial handler
    th = exp.create_trialhandler()
    assert th.trialList == [default_trial(), default_trial()]
    assert th.extraInfo["metadata"]["author"] == "Joe Smith"
    assert th.extraInfo["display_options"]["to_target_paths"] is False
    assert th.finished is False
    # save the experiment (psydat file extension is automatically added if missing)
    filename_to_save = str(tmp_path / "ex1")
    filename_to_load = str(tmp_path / "ex1.psydat")
    exp.save_psydat(filename_to_save)
    assert exp.has_unsaved_changes is False
    # load the experiment
    exp2 = Experiment(filename_to_load)
    assert exp2.metadata["author"] == "Joe Smith"
    assert exp2.display_options["to_target_paths"] is False
    assert exp2.has_unsaved_changes is False
    assert exp.trial_handler_with_results is None
    # get a new trial handler
    th2 = exp2.create_trialhandler()
    assert th2.trialList == [default_trial(), default_trial()]
    assert th2.extraInfo["metadata"]["author"] == "Joe Smith"
    assert th2.extraInfo["display_options"]["to_target_paths"] is False
    assert th2.finished is False


def test_experiment_with_results(
    experiment_with_results: Experiment, tmp_path: pathlib.Path
) -> None:
    # initial experiment has existing results
    assert experiment_with_results.trial_handler_with_results is not None
    assert experiment_with_results.trial_handler_with_results.finished
    assert (
        "to_target_timestamps"
        in experiment_with_results.trial_handler_with_results.data
    )
    assert experiment_with_results.has_unsaved_changes is True
    # get a new trial handler for this experiment
    th = experiment_with_results.create_trialhandler()
    assert th.trialList == experiment_with_results.trial_list
    assert th.extraInfo["metadata"] == experiment_with_results.metadata
    assert th.extraInfo["display_options"] == experiment_with_results.display_options
    assert "to_target_timestamps" not in th.data
    assert th.finished is False
    # save experiment
    filename = str(tmp_path / "ex1.psydat")
    experiment_with_results.save_psydat(filename)
    assert experiment_with_results.has_unsaved_changes is False
    # load the experiment
    exp2 = Experiment(filename)
    assert exp2.metadata == experiment_with_results.metadata
    assert exp2.display_options == experiment_with_results.display_options
    assert exp2.trial_list == experiment_with_results.trial_list
    assert exp2.has_unsaved_changes is False
    assert exp2.trial_handler_with_results is not None
    for key in [
        "target_pos",
        "to_target_timestamps",
        "to_center_timestamps",
        "to_target_mouse_positions",
        "to_center_mouse_positions",
    ]:
        for trial1, trial2 in zip(
            exp2.trial_handler_with_results.data[key],
            experiment_with_results.trial_handler_with_results.data[key],
        ):
            for rep1, rep2 in zip(trial1, trial2):
                for data1, data2 in zip(rep1, rep2):
                    assert np.allclose(data1, data2)
    # get a new trial handler
    th2 = exp2.create_trialhandler()
    assert th2.trialList == experiment_with_results.trial_list
    assert th2.extraInfo["metadata"] == experiment_with_results.metadata
    assert th2.extraInfo["display_options"] == experiment_with_results.display_options
    assert "to_target_timestamps" not in th2.data
    assert th2.finished is False


def test_experiment_to_excel_target_data_format(
    experiment_with_results: Experiment, tmp_path: pathlib.Path
) -> None:
    # export to an Excel file with a sheet for each target
    excel_file = tmp_path / "export.xlsx"
    experiment_with_results.save_excel(str(excel_file), data_format="target")
    stats: pd.DataFrame = experiment_with_results.stats
    # import this Excel file again
    exp = Experiment(str(excel_file))
    assert exp.metadata == experiment_with_results.metadata
    assert exp.display_options == experiment_with_results.display_options
    assert exp.trial_list == experiment_with_results.trial_list
    assert exp.trial_handler_with_results is None
    assert exp.has_unsaved_changes is True
    assert exp.filename == str(excel_file.with_suffix(".psydat"))
    # check that results sheets in Excel are consistent with stats dataframe
    dfs = pd.read_excel(excel_file, sheet_name=None)
    df_stats = dfs["statistics"]
    # convert x,y columns back to single (x,y) tuple column
    df_stats["center_pos"] = list(zip(df_stats.center_pos_x, df_stats.center_pos_y))
    df_stats.drop(columns=["center_pos_x", "center_pos_y"], inplace=True)
    df_stats["target_pos"] = list(zip(df_stats.target_pos_x, df_stats.target_pos_y))
    df_stats.drop(columns=["target_pos_x", "target_pos_y"], inplace=True)
    for name in df_stats:
        for a, b in zip(stats[name], df_stats[name]):
            assert np.allclose(a, b, equal_nan=True)
    assert len(dfs) == 4 + sum(
        [
            trial["weight"] * trial["num_targets"]
            for trial in experiment_with_results.trial_list
        ]
    )
    # check that excel sheet for each target is consistent with corresponding stats dataframe row
    for i_row in range(stats.shape[0]):
        df_data = dfs[f"{i_row}"]
        for label in ["to_target_timestamps", "to_center_timestamps"]:
            correct_values = stats[label][i_row]
            n = correct_values.shape[0]
            assert np.allclose(df_data[label].to_numpy()[0:n], correct_values)
        for label in ["to_target_mouse_positions", "to_center_mouse_positions"]:
            correct_values = stats[label][i_row]
            n = correct_values.shape[0]
            if n > 0:
                assert np.allclose(
                    df_data[f"{label}_x"].to_numpy()[0:n], correct_values[:, 0]
                )
                assert np.allclose(
                    df_data[f"{label}_y"].to_numpy()[0:n], correct_values[:, 1]
                )


def test_experiment_to_excel_trial_data_format(
    experiment_with_results: Experiment, tmp_path: pathlib.Path
) -> None:
    # export to an Excel file with a sheet for each trial
    excel_file = tmp_path / "export.xlsx"
    experiment_with_results.save_excel(str(excel_file), data_format="trial")
    stats: pd.DataFrame = experiment_with_results.stats
    # import this Excel file again
    exp = Experiment(str(excel_file))
    assert exp.metadata == experiment_with_results.metadata
    assert exp.display_options == experiment_with_results.display_options
    assert exp.trial_list == experiment_with_results.trial_list
    assert exp.trial_handler_with_results is None
    assert exp.has_unsaved_changes is True
    assert exp.filename == str(excel_file.with_suffix(".psydat"))
    # check that results sheets in Excel are consistent with stats dataframe
    dfs = pd.read_excel(excel_file, sheet_name=None)
    df_stats = dfs["statistics"]
    # convert x,y columns back to single (x,y) tuple column
    df_stats["center_pos"] = list(zip(df_stats.center_pos_x, df_stats.center_pos_y))
    df_stats.drop(columns=["center_pos_x", "center_pos_y"], inplace=True)
    df_stats["target_pos"] = list(zip(df_stats.target_pos_x, df_stats.target_pos_y))
    df_stats.drop(columns=["target_pos_x", "target_pos_y"], inplace=True)
    for name in df_stats:
        for a, b in zip(stats[name], df_stats[name]):
            assert np.allclose(a, b, equal_nan=True)
    n_trials = sum([trial["weight"] for trial in experiment_with_results.trial_list])
    assert len(dfs) == 4 + n_trials
    # check that excel trial data is consistent with stats dataframe
    n_center_targets_to_check = sum(
        [
            trial["weight"]
            * trial["num_targets"]
            * (0 if trial["automove_cursor_to_center"] else 1)
            for trial in experiment_with_results.trial_list
        ]
    )
    n_center_targets_checked = 0
    n_outer_targets_to_check = sum(
        [
            trial["weight"] * trial["num_targets"]
            for trial in experiment_with_results.trial_list
        ]
    )
    n_outer_targets_checked = 0
    i_row = -1  # track which row of df we should compare with
    for i_trial in range(n_trials):
        # this contains data for all targets from this trial
        df_data = dfs[f"{i_trial}"]
        # find the locations where the target changes
        indices_where_target_changes = df_data.index[
            df_data.i_target.diff() != 0
        ].values.tolist()
        indices_where_target_changes.append(df_data.index[-1] + 1)
        # iterate over each chunk of data with a single target
        for start, end in zip(
            indices_where_target_changes, indices_where_target_changes[1:]
        ):
            df_target = df_data.loc[start : end - 1]
            i_target = df_target.i_target.iloc[0]
            if i_target >= 0:
                dest = "target"
                # increment row counter each time we see a new (outer) target in the excel data
                i_row += 1
            elif i_target == -1:
                dest = "center"
            else:
                dest = None
            if dest is not None:
                # target data from target display time
                # get correct timestamps
                ts_correct = stats[f"to_{dest}_timestamps"][i_row]
                # get imported timestamps
                ts = df_target.timestamps.to_numpy()
                assert np.allclose(ts, ts_correct)
                # get correct mouse positions
                xys = stats[f"to_{dest}_mouse_positions"][i_row]
                n = xys.shape[0]
                if dest == "target" or (dest == "center" and n > 0):
                    assert np.allclose(
                        df_target["mouse_positions_x"].to_numpy(), xys[:, 0]
                    )
                    assert np.allclose(
                        df_target["mouse_positions_y"].to_numpy(), xys[:, 1]
                    )
                    if dest == "target":
                        n_outer_targets_checked += 1
                    elif dest == "center":
                        n_center_targets_checked += 1
            else:
                # data from where no target is visible: should also test this!
                pass
    assert n_outer_targets_checked == n_outer_targets_to_check
    assert n_center_targets_checked == n_center_targets_to_check


def test_experiment_to_excel_invalid_data_format(
    experiment_with_results: Experiment, tmp_path: pathlib.Path
) -> None:
    excel_file = tmp_path / "export.xlsx"
    with pytest.raises(RuntimeError):
        experiment_with_results.save_excel(str(excel_file), data_format="invalid")


def test_experiment_to_json(
    experiment_with_results: Experiment, tmp_path: pathlib.Path
) -> None:
    # export to a json file
    json_file = tmp_path / "export.json"
    experiment_with_results.save_json(str(json_file))
    # import this json file again
    exp = Experiment(str(json_file))
    assert exp.metadata == experiment_with_results.metadata
    assert exp.display_options == experiment_with_results.display_options
    assert exp.trial_list == experiment_with_results.trial_list
    assert exp.trial_handler_with_results is None
    assert exp.has_unsaved_changes is True
    assert exp.filename == str(json_file.with_suffix(".psydat"))


def test_experiment_invalid_file(tmp_path: pathlib.Path) -> None:
    # file doesn't exist
    with pytest.raises(ValueError):
        Experiment("I don't exist!")

    # file exists and has psydat extension but is garbage
    file = tmp_path / "invalid.psydat"
    file.write_text("beep boop")
    with pytest.raises(pickle.PickleError):
        Experiment(str(file))

    # file exists and has xlsx extension but is garbage
    file = tmp_path / "invalid.xlsx"
    file.write_text("beep boop")
    with pytest.raises(ValueError):
        Experiment(str(file))


def test_experiment_empty_trialhandler() -> None:
    # construct an empty TrialHandlerExt: no trials, no extraInfo dict
    empty_th = TrialHandlerExt([], 1)
    # this is what gets constructed from an empty trial list:
    assert empty_th.trialList == [None]
    assert empty_th.extraInfo is None
    experiment = Experiment()
    experiment.import_and_validate_trial_handler(empty_th)
    # defaults are used for missing metadata and display_options
    assert experiment.metadata == default_metadata()
    assert experiment.display_options == default_display_options()
    assert len(experiment.trial_list) == 0



from __future__ import annotations

import logging
import subprocess
import sys

import requests
from packaging import version

import vstt


def _get_latest_pypi_version() -> version.Version:
    logging.info("Requesting latest version from pypi...")
    r = requests.get("https://pypi.org/pypi/vstt/json", timeout=5)
    version_str = r.json().get("info", {}).get("version", "")
    ver = version.parse(version_str)
    logging.info(f"  -> {ver}")
    return ver


def check_for_new_version() -> tuple[bool, str]:
    try:
        current_version = version.Version(vstt.__version__)
        logging.info(f"Current version: {current_version}")
        latest_version = _get_latest_pypi_version()
        logging.info(f"Latest version: {latest_version}")
        if latest_version > current_version:
            return (
                True,
                f"Latest version of VSTT: {latest_version}.\nYou currently have version {current_version}.\nWould you like to upgrade now?",
            )
        else:
            return (
                False,
                "You have the latest version of VSTT.",
            )
    except Exception as e:
        logging.exception(e)
        return (
            False,
            "An error occurred while checking for updates, please try again later.",
        )


def do_pip_upgrade() -> tuple[bool, str]:
    logging.info(f"Doing pip upgrade using {sys.executable}...")
    try:
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "--upgrade", "vstt"]
        )
        logging.info("done.")
        return (
            True,
            "VSTT has been updated.\nPlease close the program and open it again to see the latest version.",
        )
    except Exception as e:
        logging.exception(e)
        return False, f"An error occurred when trying to update VSTT: {e}"



from __future__ import annotations

import gui_test_utils as gtu
import qt_test_utils as qtu
from psychopy.visual.window import Window
from pytest import approx
from qtpy import QtWidgets

import vstt
from vstt.experiment import Experiment
from vstt.meta_widget import MetadataWidget


def test_metadata_widget(window: Window) -> None:
    widget = MetadataWidget(parent=None, win=window)
    signal_received = qtu.SignalReceived(widget.experiment_modified)
    # default-constructed widget has a default experiment with default metadata
    assert widget.experiment.metadata == vstt.meta.default_metadata()
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    screenshot = gtu.call_target_and_get_screenshot(
        qtu.click,
        (widget._btn_preview_metadata,),
        window,
    )
    # most pixels are grey
    default_grey_pixel_fraction = gtu.pixel_color_fraction(screenshot, (128, 128, 128))
    assert 0.850 < default_grey_pixel_fraction < 0.999
    # a couple percent of black pixels due to black text
    assert 0.005 <= gtu.pixel_color_fraction(screenshot, (0, 0, 0)) <= 0.050
    experiment = Experiment()
    empty_metadata = vstt.meta.default_metadata()
    for key, value in empty_metadata.items():
        if isinstance(value, str):
            empty_metadata[key] = ""  # type: ignore
    empty_metadata["show_delay_countdown"] = False
    experiment.metadata = empty_metadata
    # assign experiment with empty metadata strings
    widget.experiment = experiment
    assert widget.experiment is experiment
    assert widget.experiment.metadata == empty_metadata
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    screenshot = gtu.call_target_and_get_screenshot(
        qtu.click,
        (widget._btn_preview_metadata,),
        window,
    )
    # all pixels grey except for blue continue text
    empty_grey_pixel_fraction = gtu.pixel_color_fraction(screenshot, (128, 128, 128))
    assert empty_grey_pixel_fraction > default_grey_pixel_fraction
    # no other text so no black pixels
    assert gtu.pixel_color_fraction(screenshot, (0, 0, 0)) == approx(0)
    # reset to experiment with empty metadata, then type variable name in each line edit
    assert widget.experiment.has_unsaved_changes is False
    assert widget.experiment.metadata == empty_metadata
    assert not signal_received
    for key in widget._str_widgets:
        line_edit = widget._str_widgets[key]
        assert line_edit is not None
        assert isinstance(line_edit, QtWidgets.QLineEdit)
        assert widget.experiment.metadata[key] == ""  # type: ignore
        signal_received.clear()
        qtu.press_keys(line_edit, key)
        assert widget.experiment.metadata[key] == key  # type: ignore
        assert widget.experiment.has_unsaved_changes is True
        assert signal_received
    for key, value in widget.experiment.metadata.items():
        if isinstance(value, str):
            assert value == key
    # assign another experiment to widget & update fields
    experiment2 = Experiment()
    experiment2.has_unsaved_changes = False
    widget.experiment = experiment2
    assert widget.experiment is experiment2
    assert widget.experiment.has_unsaved_changes is False
    for key, value in vstt.meta.default_metadata().items():
        if isinstance(value, str):
            line_edit = widget._str_widgets[key]
            signal_received.clear()
            assert line_edit is not None
            assert type(line_edit) is QtWidgets.QLineEdit
            assert widget.experiment.metadata[key] == value  # type: ignore
            qtu.press_keys(line_edit, "2")
            assert widget.experiment.metadata[key] == value + "2"  # type: ignore
            assert widget.experiment.has_unsaved_changes is True
            assert signal_received
    # experiment2 has been updated
    for key, value in experiment2.metadata.items():
        if isinstance(value, str):
            assert value == vstt.meta.default_metadata()[key] + "2"  # type: ignore
    # previous experiment was not modified
    for key, value in experiment.metadata.items():
        if isinstance(value, str):
            assert value == key



from __future__ import annotations

import gui_test_utils as gtu
import numpy as np
import pytest
from psychopy.visual.window import Window
from pytest import approx

import vstt
from vstt.experiment import Experiment


def test_make_cursor(window: Window) -> None:
    for cursor_size in [0.01, 0.02235457]:
        cursor = vstt.vis.make_cursor(window, cursor_size)
        assert np.allclose(cursor.pos, [0, 0])
        # x size:
        assert np.allclose(cursor.vertices[1][0] - cursor.vertices[0][0], cursor_size)
        # y size:
        assert np.allclose(cursor.vertices[-1][1] - cursor.vertices[-2][1], cursor_size)


@pytest.mark.parametrize("add_central_target", [True, False])
@pytest.mark.parametrize(
    "args,xys",
    [
        (
            {
                "n_circles": 1,
                "radius": 0.5,
                "point_radius": 0.1,
                "center_point_radius": 0.05,
            },
            [(0, 0.5), (0, 0)],
        ),
        (
            {
                "n_circles": 2,
                "radius": 0.3,
                "point_radius": 0.02,
                "center_point_radius": 0.04,
            },
            [(0, 0.3), (0, -0.3), (0, 0)],
        ),
        (
            {
                "n_circles": 4,
                "radius": 0.8,
                "point_radius": 0.2,
                "center_point_radius": 0.07,
            },
            [(0, 0.8), (0.8, 0), (0, -0.8), (-0.8, 0), (0, 0)],
        ),
    ],
)
def test_make_targets(
    window: Window, args: dict, xys: list[tuple[float, float]], add_central_target: bool
) -> None:
    args["add_central_target"] = add_central_target
    targets = vstt.vis.make_targets(window, **args)
    n_elem = args["n_circles"] + (1 if add_central_target else 0)
    assert targets.nElements == n_elem
    # shape of sizes is x,y pair for each element
    assert targets.sizes.shape == (n_elem, 2)
    # size is circumference of circle in both x and y directions
    point_size = 2 * args["point_radius"]
    for size in targets.sizes[: args["n_circles"]]:
        assert np.allclose(size, [point_size, point_size])
    if add_central_target:
        center_point_size = 2 * args["center_point_radius"]
        assert np.allclose(targets.sizes[-1], [center_point_size, center_point_size])
    assert np.allclose(targets.xys, xys[0:n_elem])


@pytest.mark.parametrize("add_central_target", [True, False])
@pytest.mark.parametrize("n_targets", [1, 2, 3, 5, 12])
def test_update_target_colors(
    window: Window, n_targets: int, add_central_target: bool
) -> None:
    for show_inactive_targets, inactive_color in [
        (True, (0.9, 0.9, 0.9)),
        (False, (0, 0, 0)),
    ]:
        red = (1.0, -1.0, -1.0)
        targets = vstt.vis.make_targets(
            window, n_targets, 0.5, 0.05, add_central_target, 0.05
        )
        n_elem = n_targets + (1 if add_central_target else 0)
        # calling without specifying an index makes all elements grey
        vstt.vis.update_target_colors(
            targets, show_inactive_targets=show_inactive_targets
        )
        if n_elem == 1:
            # apparently psychopy collapses (n_elem,3) shape to (3) for n_elem=1 case:
            assert targets.colors.shape == (3,)
        else:
            assert targets.colors.shape == (n_elem, 3)
        for color in targets.colors:
            assert np.allclose(color, inactive_color)
        # calling with index makes that element red, the rest grey
        for index in range(n_elem):
            vstt.vis.update_target_colors(
                targets, show_inactive_targets=show_inactive_targets, index=index
            )
            if n_elem == 1:
                # apparently psychopy collapses (n_elem,3) shape to (3) for n_elem=1 case:
                assert targets.colors.shape == (3,)
            else:
                assert targets.colors.shape == (n_elem, 3)
            for i in range(n_elem):
                if i == index:
                    if n_elem == 1:
                        # apparently psychopy collapses (n_elem,3) shape to (3) for n_elem=1 case:
                        assert np.allclose(targets.colors, red)
                    else:
                        assert np.allclose(targets.colors[i], red)
                else:
                    assert np.allclose(targets.colors[i], inactive_color)


@pytest.mark.parametrize(
    "args,xys",
    [
        (
            {
                "n_circles": 1,
                "radius": 0.44,
                "point_radius": 0.1,
            },
            [(0, 0.44)],
        ),
        (
            {
                "n_circles": 4,
                "radius": 0.3,
                "point_radius": 0.05,
            },
            [(0, 0.3), (0.3, 0), (0, -0.3), (-0.3, 0)],
        ),
    ],
)
def test_make_target_labels(
    window: Window, args: dict, xys: list[tuple[float, float]]
) -> None:
    # more labels than targets: label each target and ignore any extra labels
    args["labels_string"] = "a b c d e f"
    label_strings = ["a", "b", "c", "d", "e", "f"]
    target_labels = vstt.vis.make_target_labels(window, **args)
    assert len(target_labels) == args["n_circles"]
    for target_label, label_string, xy in zip(target_labels, label_strings, xys):
        assert target_label.text == label_string
        assert target_label.letterHeight < 2 * args["point_radius"]
        assert np.allclose(target_label.pos, xy)
    # equal or fewer labels than targets: just use available labels
    args["labels_string"] = "10"
    label_strings = ["10"]
    target_labels = vstt.vis.make_target_labels(window, **args)
    assert len(target_labels) == 1
    for target_label, label_string, xy in zip(target_labels, label_strings, xys):
        assert target_label.text == label_string
        assert target_label.letterHeight < 2 * args["point_radius"]
        assert np.allclose(target_label.pos, xy)


@pytest.mark.parametrize("n_targets", [1, 2, 3, 5, 12])
def test_update_target_label_colors(window: Window, n_targets: int) -> None:
    labels = "a b c d e f g h i j k l m n o p"
    for show_inactive_targets, inactive_color in [
        (True, (0.3, 0.3, 0.3)),
        (False, (0, 0, 0)),
    ]:
        white = (1.0, 1.0, 1.0)
        target_labels = vstt.vis.make_target_labels(
            window, n_targets, 0.5, 0.05, labels
        )
        # calling without specifying an index makes all elements grey
        vstt.vis.update_target_label_colors(
            target_labels, show_inactive_targets=show_inactive_targets
        )
        assert len(target_labels) == n_targets
        for target_label in target_labels:
            assert np.allclose(target_label.color, inactive_color)
        # calling with index makes that element red, the rest grey
        for index in range(n_targets):
            vstt.vis.update_target_label_colors(
                target_labels, show_inactive_targets=show_inactive_targets, index=index
            )
            assert len(target_labels) == n_targets
            for i, target_label in enumerate(target_labels):
                if i == index:
                    assert np.allclose(target_label.color, white)
                else:
                    assert np.allclose(target_label.color, inactive_color)


def test_splash_screen_defaults(window: Window) -> None:
    metadata = vstt.meta.default_metadata()
    screenshot = gtu.call_target_and_get_screenshot(
        vstt.vis.splash_screen, (1000, True, False, metadata, window), window
    )
    # most pixels grey except for black main text and blue continue text
    assert 0.900 < gtu.pixel_color_fraction(screenshot, (128, 128, 128)) < 0.999
    # some black pixels
    assert 0.000 < gtu.pixel_color_fraction(screenshot, (0, 0, 0)) < 0.100
    # no white pixels
    assert gtu.pixel_color_fraction(screenshot, (255, 255, 255)) == approx(0)


def test_display_results_nothing(
    experiment_with_results: Experiment, window: Window
) -> None:
    experiment_with_results.display_options = {
        "to_target_paths": False,
        "to_center_paths": False,
        "targets": False,
        "central_target": False,
        "to_target_reaction_time": False,
        "to_center_reaction_time": False,
        "to_target_movement_time": False,
        "to_center_movement_time": False,
        "to_target_time": False,
        "to_center_time": False,
        "to_target_distance": False,
        "to_center_distance": False,
        "to_target_rmse": False,
        "to_center_rmse": False,
        "averages": False,
        "to_target_success": False,
        "to_center_success": False,
        "area": False,
        "normalized_area": False,
        "peak_velocity": False,
        "peak_acceleration": False,
        "to_target_spatial_error": False,
        "to_center_spatial_error": False,
        "movement_time_at_peak_velocity": False,
        "total_time_at_peak_velocity": False,
        "movement_distance_at_peak_velocity": False,
        "rmse_movement_at_peak_velocity": False,
    }
    for all_trials_for_this_condition in [False, True]:
        # trial 0: 0,1,2 are trials without auto-move to center
        screenshot = gtu.call_target_and_get_screenshot(
            vstt.vis.display_results,
            (
                60,
                True,
                False,
                experiment_with_results.trial_handler_with_results,
                experiment_with_results.display_options,
                0,
                all_trials_for_this_condition,
                window,
            ),
            window,
        )
        # all pixels grey except for blue continue text
        assert 0.990 < gtu.pixel_color_fraction(screenshot, (128, 128, 128)) < 0.999
        # no off-white pixels
        assert gtu.pixel_color_fraction(screenshot, (240, 248, 255)) == approx(0)
        # trial 3: with auto-move to center
        screenshot = gtu.call_target_and_get_screenshot(
            vstt.vis.display_results,
            (
                60,
                True,
                False,
                experiment_with_results.trial_handler_with_results,
                experiment_with_results.display_options,
                3,
                all_trials_for_this_condition,
                window,
            ),
            window,
        )
        # all pixels grey except for blue continue text
        assert 0.990 < gtu.pixel_color_fraction(screenshot, (128, 128, 128)) < 0.999
        # no off-white pixels
        assert gtu.pixel_color_fraction(screenshot, (240, 248, 255)) == approx(0)


def test_display_results_everything(
    experiment_with_results: Experiment, window: Window
) -> None:
    experiment_with_results.display_options = {
        "to_target_paths": True,
        "to_center_paths": True,
        "targets": True,
        "central_target": True,
        "to_target_reaction_time": True,
        "to_center_reaction_time": True,
        "to_target_movement_time": True,
        "to_center_movement_time": True,
        "to_target_time": True,
        "to_center_time": True,
        "to_target_distance": True,
        "to_center_distance": True,
        "to_target_rmse": True,
        "to_center_rmse": True,
        "averages": True,
        "to_target_success": True,
        "to_center_success": True,
        "area": True,
        "normalized_area": True,
        "peak_velocity": True,
        "peak_acceleration": True,
        "to_target_spatial_error": True,
        "to_center_spatial_error": True,
        "movement_time_at_peak_velocity": True,
        "total_time_at_peak_velocity": True,
        "movement_distance_at_peak_velocity": True,
        "rmse_movement_at_peak_velocity": True,
    }
    for all_trials_for_this_condition in [False, True]:
        # trial 0: 0,1,2 are trials without auto-move to center
        screenshot = gtu.call_target_and_get_screenshot(
            vstt.vis.display_results,
            (
                60,
                True,
                False,
                experiment_with_results.trial_handler_with_results,
                experiment_with_results.display_options,
                0,
                all_trials_for_this_condition,
                window,
            ),
            window,
        )
        # less grey: lots of other colors for targets, paths and stats
        grey_pixels = gtu.pixel_color_fraction(screenshot, (128, 128, 128))
        assert 0.750 < grey_pixels < 0.950
        # some off-white pixels for one of the targets
        off_white_pixels = gtu.pixel_color_fraction(screenshot, (240, 248, 255))
        assert 0.002 < off_white_pixels < 0.100
        # trial 3: with auto-move to center
        screenshot = gtu.call_target_and_get_screenshot(
            vstt.vis.display_results,
            (
                60,
                True,
                False,
                experiment_with_results.trial_handler_with_results,
                experiment_with_results.display_options,
                3,
                all_trials_for_this_condition,
                window,
            ),
            window,
        )
        # more grey since there are fewer paths and stats to display
        assert gtu.pixel_color_fraction(screenshot, (128, 128, 128)) > grey_pixels



from __future__ import annotations

import vstt


def test_metadata_labels() -> None:
    metadata = vstt.meta.default_metadata()
    labels = vstt.meta.metadata_labels()
    assert len(metadata) == len(labels)
    assert metadata.keys() == labels.keys()


def test_import_metadata() -> None:
    # valid metadata imported
    valid_dict = {
        "name": "bob",
        "subject": "trial",
        "date": "1/2/3",
        "author": "jim",
        "display_title": "title",
        "display_text1": "t1",
        "display_text2": "t2",
        "display_text3": "t3",
        "display_text4": "t4",
        "display_duration": 123,
        "show_delay_countdown": False,
        "enter_to_skip_delay": False,
    }
    metadata = vstt.meta.import_metadata(valid_dict)
    assert metadata == valid_dict
    # all keys missing -> replaced with defaults
    metadata = vstt.meta.import_metadata({})
    assert metadata == vstt.meta.default_metadata()
    # valid keys present but values have wrong type:
    #   - coerced to correct type if legal to do so
    #   - otherwise replaced with defaults
    metadata = vstt.meta.import_metadata(
        {"name": 12, "data": 2.4, "author": [1, 2, 3], "whoops": "ok"}
    )
    assert metadata["name"] == "12"
    assert metadata["author"] == "[1, 2, 3]"
    # unknown keys are ignored
    assert "data" not in metadata
    assert "whoops" not in metadata



"""
Wrapper around psychopy Joystick to work around psychopy bug

For pyglet backend psychopy Joystick constructor calls .open() on the pyglet input device.
If this device is already open (i.e. we already created a Joystick) then pyglet raises an exception.
Here we keep track of the current joystick object and attempt to close the underlying pyglet device.
"""

from __future__ import annotations

import logging

from psychopy.hardware import joystick
from pyglet.input import DeviceOpenException

js: joystick.Joystick | None = None


def _close_open_device() -> None:
    global js
    if js is None:
        return
    try:
        js._device.close()  # attempt to close pyglet device
    except Exception as e:
        logging.warning(f"Failed to close pyglet joystick device: {e}")
        logging.exception(e)


def _updated_joystick() -> joystick.Joystick:
    try:
        return joystick.Joystick(0)
    except DeviceOpenException:
        _close_open_device()
        return joystick.Joystick(0)


def get_joystick() -> joystick.Joystick | None:
    global js
    if joystick.getNumJoysticks() == 0:
        return None
    try:
        js = _updated_joystick()
    except Exception as e:
        logging.warning("Failed to get Joystick")
        logging.exception(e)
        return None
    return js



from __future__ import annotations

import json
import pathlib
import pickle
from typing import Any

import pandas as pd
from psychopy.data import TrialHandlerExt
from psychopy.misc import fromFile

from vstt.display import default_display_options
from vstt.display import import_display_options
from vstt.meta import default_metadata
from vstt.meta import import_metadata
from vstt.stats import append_stats_data_to_excel
from vstt.stats import stats_dataframe
from vstt.trial import default_trial
from vstt.trial import import_and_validate_trial


class Experiment:
    def __init__(self, filename: str | None = None):
        self.filename = "default-experiment.psydat"
        self.has_unsaved_changes = False
        self.metadata = default_metadata()
        self.display_options = default_display_options()
        self.trial_list = [default_trial()]
        self.trial_handler_with_results: TrialHandlerExt | None = None
        self.stats: pd.DataFrame | None = None
        if filename is not None:
            self.load_file(filename)

    def create_trialhandler(self) -> TrialHandlerExt:
        for index, trial in enumerate(self.trial_list):
            self.trial_list[index] = import_and_validate_trial(trial)
        return TrialHandlerExt(
            self.trial_list,
            nReps=1,
            method="sequential",
            originPath=-1,
            extraInfo={
                "display_options": self.display_options,
                "metadata": self.metadata,
            },
        )

    def clear_results(self) -> None:
        self.trial_handler_with_results = None

    def _as_dict(self) -> dict[str, Any]:
        return {
            "metadata": self.metadata,
            "display_options": self.display_options,
            "trial_list": self.trial_list,
        }

    def load_file(self, filename: str) -> None:
        suffix = pathlib.Path(filename).suffix
        if suffix == ".xlsx":
            self.load_excel(filename)
        elif suffix == ".json":
            self.load_json(filename)
        else:
            # assume psydat format by default
            self.load_psydat(filename)

    def save_psydat(self, filename: str) -> None:
        if not filename.endswith(".psydat"):
            filename += ".psydat"

        if self.trial_handler_with_results is not None:
            # save existing trial handler with results, only update its extraInfo dict
            self.trial_handler_with_results.extraInfo = {
                "display_options": self.display_options,
                "metadata": self.metadata,
            }
            self.trial_handler_with_results.saveAsPickle(
                filename, fileCollisionMethod="overwrite"
            )
        else:
            # create a new trial handler to save this experiment
            trial_handler = self.create_trialhandler()
            # temporary hack as the psyschopy function only saves if there are results!
            with open(filename, "wb") as f:
                pickle.dump(trial_handler, f)
        self.filename = filename
        self.has_unsaved_changes = False

    def load_psydat(self, filename: str) -> None:
        self.import_and_validate_trial_handler(fromFile(filename))
        self.filename = filename
        self.has_unsaved_changes = False

    def save_excel(self, filename: str, data_format: str) -> None:
        """
        data_format can be
        - "trial": one sheet of data exported per trial
        - "target: one sheet of data exported per target
        """
        if not filename.endswith(".xlsx"):
            filename += ".xlsx"
        with pd.ExcelWriter(filename) as writer:
            pd.DataFrame([self.metadata]).to_excel(
                writer, sheet_name="metadata", index=False
            )
            pd.DataFrame([self.display_options]).to_excel(
                writer, sheet_name="display_options", index=False
            )
            pd.DataFrame(self.trial_list).to_excel(
                writer, sheet_name="trial_list", index=False
            )
            if self.stats is not None:
                append_stats_data_to_excel(self.stats, writer, data_format)

    def load_excel(self, filename: str) -> None:
        dfs = pd.read_excel(filename, ["metadata", "display_options", "trial_list"])
        self.import_and_validate_dicts(
            filename,
            dfs["metadata"].to_dict("records")[0],
            dfs["display_options"].to_dict("records")[0],
            dfs["trial_list"].to_dict("records"),
        )

    def save_json(self, filename: str) -> None:
        if not filename.endswith(".json"):
            filename += ".json"
        with open(filename, "w") as f:
            json.dump(self._as_dict(), f)

    def load_json(self, filename: str) -> None:
        with open(filename) as f:
            d = json.load(f)
        self.import_and_validate_dicts(
            filename, d["metadata"], d["display_options"], d["trial_list"]
        )

    def import_and_validate_dicts(
        self,
        filename: str,
        metadata_dict: dict,
        display_options_dict: dict,
        trial_dict_list: list[dict],
    ) -> None:
        self.metadata = import_metadata(metadata_dict)
        self.display_options = import_display_options(display_options_dict)
        self.trial_list = [
            import_and_validate_trial(trial_dict) for trial_dict in trial_dict_list
        ]
        self.trial_handler_with_results = None
        self.stats = None
        self.has_unsaved_changes = True
        self.filename = str(pathlib.Path(filename).with_suffix(".psydat"))

    def import_and_validate_trial_handler(self, trial_handler: TrialHandlerExt) -> None:
        # psychopy trial handler converts empty trial list [] -> [None]
        if trial_handler.trialList == [None]:
            trial_handler.trialList = []
        for index, trial in enumerate(trial_handler.trialList):
            trial_handler.trialList[index] = import_and_validate_trial(trial)
        self.trial_list = trial_handler.trialList
        if not trial_handler.extraInfo:
            trial_handler.extraInfo = {}
        self.metadata = import_metadata(
            trial_handler.extraInfo.get("metadata", default_metadata())
        )
        self.display_options = import_display_options(
            trial_handler.extraInfo.get("display_options", default_display_options())
        )
        if trial_handler.finished:
            self.trial_handler_with_results = trial_handler
            self.stats = stats_dataframe(trial_handler)
        else:
            self.trial_handler_with_results = None
            self.stats = None



from __future__ import annotations

import numpy as np
from pytest import approx

import vstt


def test_equidistant_angles() -> None:
    assert np.allclose(vstt.geom.equidistant_angles(1) / np.pi, [0])
    assert np.allclose(vstt.geom.equidistant_angles(2) / np.pi, [0, 1])
    assert np.allclose(
        vstt.geom.equidistant_angles(3) / np.pi, [0, 2.0 / 3.0, 4.0 / 3.0]
    )
    assert np.allclose(vstt.geom.equidistant_angles(4) / np.pi, [0, 0.5, 1.0, 1.5])


def test_points_on_circle() -> None:
    assert np.allclose(vstt.geom.points_on_circle(1, 0.2), [(0.0, 0.2)])
    assert np.allclose(
        vstt.geom.points_on_circle(1, 0.2, include_centre=True),
        [(0.0, 0.2), (0.0, 0.0)],
    )
    assert np.allclose(vstt.geom.points_on_circle(2, 0.5), [(0.0, 0.5), (0.0, -0.5)])
    assert np.allclose(
        vstt.geom.points_on_circle(4, 1.0), [(0, 1), (1, 0), (0, -1), (-1, 0)]
    )


def test_rotate_point() -> None:
    pivots = [(0, 0), (0.3, 0.2), (-0.3, 0.4), (0.6, -0.4), (-1, -1)]
    points = [(1, 0.2), (-1, 0.2), (-0.7, -0.5), (0.7, 0.0), (0.0, 0.6), (0, 0)]
    angles = [0, -0.1, 0.3, 0.66, np.pi, -np.pi, 4.5, -5.2, 2.0 * np.pi]
    # rotate by zero or 360 degrees around any pivot point returns original point
    for angle in [0, -2 * np.pi, 2 * np.pi]:
        for pivot in pivots:
            for point in points:
                assert np.allclose(vstt.geom.rotate_point(point, angle, pivot), point)
    # rotating a point around itself returns the original point
    for angle in angles:
        for point in points:
            assert np.allclose(vstt.geom.rotate_point(point, angle, point), point)
    # rotating by +n degrees is equivalent to rotating by n-360 degrees
    for angle in angles:
        equivalent_angle = angle - 2 * np.pi
        for pivot in pivots:
            for point in points:
                assert np.allclose(
                    vstt.geom.rotate_point(point, angle, pivot),
                    vstt.geom.rotate_point(point, equivalent_angle, pivot),
                )


def test_point_rotator() -> None:
    for angle_radians in [0, -0.1, 0.3, 0.66, np.pi, -np.pi, 4.5]:
        angle_degrees = 180.0 * angle_radians / np.pi
        rp = vstt.geom.PointRotator(angle_degrees)
        for point in [
            (1, 0.2),
            (-1, 0.2),
            (-0.7, -0.5),
            (0.7, 0.0),
            (0.0, 0.6),
            (0, 0),
        ]:
            p1 = rp(point)
            p2 = vstt.geom.rotate_point(point, angle_radians, (0, 0))
            assert np.allclose([p1], [p2])


def test_joystick_point_updater_clipped() -> None:
    for window_size in [
        None,
        np.array([1]),
        np.array([800, 600]),
        np.array([2000, 3000]),
    ]:
        max_h = 0.5
        if window_size is not None and window_size.shape[-1] > 1:
            max_w = 0.5 * window_size[0] / window_size[1]
        else:
            max_w = 0.5
        jpu = vstt.geom.JoystickPointUpdater(
            angle_degrees=0, max_speed=0.123, window_size=window_size
        )
        for x0 in [np.array([0, 0]), np.array([0.4, -0.7]), np.array([12, 18])]:
            # default clipping if no window_size provided is to [-0.5, 0.5]
            # (note: + value joystick in y direction points down!)
            assert np.allclose(jpu(x0, (215423, 23423423)), [max_w, -max_h])
            assert np.allclose(jpu(x0, (215423, -23423423)), [max_w, max_h])
            assert np.allclose(jpu(x0, (-215423, 23423423)), [-max_w, -max_h])
            assert np.allclose(jpu(x0, (-215423, -23423423)), [-max_w, max_h])


def test_joystick_point_updater_no_clipping() -> None:
    # low max speed to avoid clipping returned values
    max_speed = 0.0001345112
    window_size = np.array([800, 600])
    for angle_radians in [0, -0.1, 0.3, 0.66, np.pi, -np.pi, 4.5]:
        angle_degrees = 180.0 * angle_radians / np.pi
        jpu = vstt.geom.JoystickPointUpdater(
            angle_degrees=angle_degrees, max_speed=max_speed, window_size=window_size
        )
        # small initial points to avoid clipping returned values
        for p0 in [
            np.array([0, 0]),
            np.array([0.002355, 0.0065334]),
            np.array([-0.00567565, 0.005464534]),
            np.array([-0.0046543235, -0.00124346344]),
        ]:
            for v in [
                (1, 0.2),
                (-1, 0.2),
                (-0.7, -0.5),
                (0.7, 0.0),
                (0.0, 0.6),
                (0, 0),
            ]:
                # invert y direction of vector to simulate joystick input (+y points down)
                p1 = jpu(p0, (v[0], -v[1]))
                # rotate vector, rescale, add to point
                p2 = p0 + max_speed * np.array(
                    vstt.geom.rotate_point((v[0], v[1]), angle_radians, (0, 0))
                )
                assert np.allclose(p1, p2)


def test_to_target_dists() -> None:
    p = np.array([0.0, 0.0])
    xys = np.array([[1.0, 0], [0.0, 0.0]])
    dist_correct, dist_any = vstt.geom.to_target_dists(
        p, xys, 0, has_central_target=True
    )
    assert dist_correct == approx(1.0)
    assert dist_any == approx(1.0)
    dist_correct, dist_any = vstt.geom.to_target_dists(
        p, xys, 0, has_central_target=False
    )
    assert dist_correct == approx(1.0)
    assert dist_any == approx(0.0)
    #
    p = np.array([1.0, 0.0])
    xys = np.array([[1.0, 1.0], [1.0, 0.0], [-1.0, 0.0], [0.0, 0.0]])
    for has_central_target in [True, False]:
        dist_correct, dist_any = vstt.geom.to_target_dists(
            p, xys, 0, has_central_target
        )
        assert dist_correct == approx(1.0)
        assert dist_any == approx(0.0)
        dist_correct, dist_any = vstt.geom.to_target_dists(
            p, xys, 1, has_central_target
        )
        assert dist_correct == approx(0.0)
        assert dist_any == approx(0.0)
        dist_correct, dist_any = vstt.geom.to_target_dists(
            p, xys, 2, has_central_target
        )
        assert dist_correct == approx(2.0)
        assert dist_any == approx(0.0)
        dist_correct, dist_any = vstt.geom.to_target_dists(
            p, xys, 3, has_central_target
        )
        assert dist_correct == approx(1.0)
        assert dist_any == approx(0.0)



from __future__ import annotations

import vstt


def test_has_valid_type() -> None:
    # type matches instance: valid
    for item in [5, 5.2, "str"]:
        assert vstt.common._has_valid_type(item, type(item))
    # float / int mismatches are also considered valid
    for number_value in [5, 5.2]:
        for number_type in [float, int]:
            assert vstt.common._has_valid_type(number_value, number_type)
    # other mismatches are invalid
    assert vstt.common._has_valid_type("5", int) is False



from __future__ import annotations

import logging
import pathlib
from typing import Callable

from psychopy.visual.window import Window
from qtpy import QtGui
from qtpy import QtWidgets
from qtpy.compat import getsavefilename
from qtpy.QtCore import Qt

import vstt
from vstt.display_widget import DisplayOptionsWidget
from vstt.experiment import Experiment
from vstt.meta_widget import MetadataWidget
from vstt.results_widget import ResultsWidget
from vstt.task import MotorTask
from vstt.trials_widget import TrialsWidget
from vstt.update import check_for_new_version
from vstt.update import do_pip_upgrade


class Gui(QtWidgets.QMainWindow):
    def __init__(self, filename: str | None = None, win: Window | None = None):
        super().__init__()
        self.experiment = Experiment()
        self._win = win

        grid_layout = QtWidgets.QVBoxLayout()
        split_top_bottom = QtWidgets.QSplitter(Qt.Vertical)

        split_metadata_display = QtWidgets.QSplitter()
        self.metadata_widget = MetadataWidget(self, win=self._win)
        self.metadata_widget.experiment_modified.connect(self.update_window_title)
        split_metadata_display.addWidget(self.metadata_widget)
        self.display_options_widget = DisplayOptionsWidget(self)
        self.display_options_widget.experiment_modified.connect(
            self.update_window_title
        )
        split_metadata_display.addWidget(self.display_options_widget)
        split_metadata_display.setSizes([1000, 1000])

        split_trial_results = QtWidgets.QSplitter()
        self.trials_widget = TrialsWidget(self)
        self.trials_widget.experiment_modified.connect(self.reload_results)
        split_trial_results.addWidget(self.trials_widget)
        self.results_widget = ResultsWidget(self, win=self._win)
        split_trial_results.addWidget(self.results_widget)

        split_top_bottom.addWidget(split_metadata_display)
        split_top_bottom.addWidget(split_trial_results)
        split_top_bottom.setSizes([3000, 1000])
        grid_layout.addWidget(split_top_bottom)

        self.toolbar = _create_menu_and_toolbar(self)
        self.addToolBar(self.toolbar)

        central_widget = QtWidgets.QWidget()
        central_widget.setLayout(grid_layout)
        self.setCentralWidget(central_widget)
        if filename is not None:
            self._open_file(filename)
        else:
            self.reload_experiment()
        self.resize(800, 600)

    def closeEvent(self, event: QtGui.QCloseEvent) -> None:
        if self.save_changes_check_continue():
            event.accept()
        else:
            event.ignore()

    def btn_new_clicked(self) -> None:
        if self.save_changes_check_continue():
            self.experiment = Experiment()
            self.reload_experiment()

    def _open_file(self, filename: str) -> None:
        try:
            self.experiment.load_file(filename)
        except Exception as e:
            logging.warning(f"Failed to load file {filename}: {e}")
            logging.exception(e)
            QtWidgets.QMessageBox.critical(
                self,
                "Invalid file",
                f"Could not read file '{filename}'",
            )
        self.reload_experiment()

    def btn_open_clicked(self) -> None:
        if self.save_changes_check_continue():
            filename, _ = QtWidgets.QFileDialog.getOpenFileName(
                self,
                "Open an experiment",
                filter="Psydat files (*.psydat) ;; Excel files (*.xlsx) ;; JSON files (*.json) ;; All files (*.*)",
                initialFilter="Psydat files (*.psydat)",
            )
            if filename == "":
                return
            self._open_file(filename)

    def btn_save_clicked(self) -> bool:
        filename, _ = getsavefilename(
            self,
            "Save experiment",
            self.experiment.filename,
            "Psydat files (*.psydat)",
        )
        if filename == "":
            return False
        try:
            self.experiment.save_psydat(filename)
        except Exception as e:
            logging.warning(f"Failed to save file {filename}: {e}")
            QtWidgets.QMessageBox.critical(
                self,
                "File save error",
                f"Could not save file '{filename}'",
            )
        self.reload_experiment()
        return True

    def btn_export_clicked(self) -> bool:
        filename, selected_filter = getsavefilename(
            self,
            "Export experiment",
            str(pathlib.Path(self.experiment.filename).with_suffix(".xlsx")),
            "Excel file (*.xlsx) ;; JSON file (*.json)",
            "Excel file (*.xlsx)",
        )
        if filename == "":
            return False
        try:
            if selected_filter == "Excel file (*.xlsx)":
                options = [
                    "One page for each trial (default)",
                    "One page for each target",
                ]
                option, ok = QtWidgets.QInputDialog.getItem(
                    self,
                    "Data export options",
                    "Data export option:",
                    options,
                    editable=False,
                )
                if ok is False:
                    return False
                data_format = "target" if "target" in option else "trial"
                self.experiment.save_excel(filename, data_format=data_format)
            elif selected_filter == "JSON file (*.json)":
                self.experiment.save_json(filename)
            else:
                raise RuntimeError(f"Selected filter {selected_filter} is not valid.")
        except Exception as e:
            logging.exception(e)
            QtWidgets.QMessageBox.critical(
                self,
                "File export error",
                f"Could not export file '{filename}'",
            )
        self.reload_experiment()
        return True

    def btn_run_clicked(self) -> None:
        if not self.experiment.trial_list:
            QtWidgets.QMessageBox.warning(
                self,
                "No trial conditions",
                "Please add a trial condition before running the experiment.",
            )
            return
        if self.experiment.trial_handler_with_results is not None:
            yes_no = QtWidgets.QMessageBox.question(
                self,
                "Clear existing results?",
                "Running the experiment will clear the existing results. Continue?",
            )
            if yes_no != QtWidgets.QMessageBox.Yes:
                return
        try:
            task = MotorTask(self.experiment, win=self._win)
            if task.run():
                self.reload_results()
        except Exception as e:
            QtWidgets.QMessageBox.warning(
                self,
                "Error running task",
                f"Error running task: {e}",
            )

    def save_changes_check_continue(self) -> bool:
        if self.experiment.has_unsaved_changes:
            yes_no = QtWidgets.QMessageBox.question(
                self,
                "Save changes?",
                "Save your changes before continuing?",
            )
            if yes_no == QtWidgets.QMessageBox.Yes:
                return self.btn_save_clicked()
        return True

    def btn_exit_clicked(self) -> None:
        self.close()

    def reload_results(self) -> None:
        self.results_widget.experiment = self.experiment
        self.update_window_title()

    def update_window_title(self) -> None:
        filename = self.experiment.filename
        if self.experiment.has_unsaved_changes:
            filename += "*"
        self.setWindowTitle(f"{filename} :: VSTT {vstt.__version__}")

    def reload_experiment(self) -> None:
        self.metadata_widget.experiment = self.experiment
        self.display_options_widget.experiment = self.experiment
        self.results_widget.experiment = self.experiment
        self.trials_widget.experiment = self.experiment
        self.update_window_title()

    def about(self) -> None:
        QtWidgets.QMessageBox.about(
            self,
            "About VSTT",
            f"Visuomotor Serial Targeting Task (VSTT)\n\nVersion {vstt.__version__}\n\n"
            + "https://ssciwr.github.io/vstt",
        )

    def check_for_updates(self) -> None:
        title = "Update VSTT"
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)
        new_version_available, new_version_message = check_for_new_version()
        QtWidgets.QApplication.restoreOverrideCursor()
        if not new_version_available:
            QtWidgets.QMessageBox.information(
                self,
                title,
                new_version_message,
            )
            return
        yes_no = QtWidgets.QMessageBox.question(self, title, new_version_message)
        if yes_no != QtWidgets.QMessageBox.Yes:
            return
        QtWidgets.QApplication.setOverrideCursor(Qt.WaitCursor)
        upgrade_success, upgrade_message = do_pip_upgrade()
        QtWidgets.QApplication.restoreOverrideCursor()
        QtWidgets.QMessageBox.information(self, title, upgrade_message)


def _add_action(
    name: str,
    callback: Callable,
    menu: QtWidgets.QMenu,
    toolbar: QtWidgets.QToolBar | None = None,
    shortcut: str | None = None,
    icon_pixmap: QtWidgets.QStyle.StandardPixmap | None = None,
) -> None:
    action = QtWidgets.QAction(name, menu)
    action.triggered.connect(callback)
    if icon_pixmap is not None:
        action.setIcon(menu.style().standardIcon(icon_pixmap))
    if shortcut is not None:
        action.setShortcut(QtGui.QKeySequence(shortcut))
    menu.addAction(action)
    if toolbar is not None:
        toolbar.addAction(action)


def _create_menu_and_toolbar(gui: vstt.gui.Gui) -> QtWidgets.QToolBar:
    menu = gui.menuBar()
    toolbar = QtWidgets.QToolBar()
    file_menu = menu.addMenu("&File")
    _add_action(
        "&New",
        gui.btn_new_clicked,
        file_menu,
        toolbar,
        "Ctrl+N",
        QtWidgets.QStyle.SP_FileIcon,
    )
    _add_action(
        "&Open",
        gui.btn_open_clicked,
        file_menu,
        toolbar,
        "Ctrl+O",
        QtWidgets.QStyle.SP_DialogOpenButton,
    )
    _add_action(
        "&Save",
        gui.btn_save_clicked,
        file_menu,
        toolbar,
        "Ctrl+S",
        QtWidgets.QStyle.SP_DialogSaveButton,
    )
    _add_action(
        "&Export",
        gui.btn_export_clicked,
        file_menu,
        toolbar,
        None,
        QtWidgets.QStyle.SP_DriveFDIcon,
    )
    _add_action(
        "E&xit",
        gui.btn_exit_clicked,
        file_menu,
        None,
        None,
        QtWidgets.QStyle.SP_DialogCloseButton,
    )
    experiment_menu = menu.addMenu("&Experiment")
    _add_action(
        "&Run",
        gui.btn_run_clicked,
        experiment_menu,
        toolbar,
        "Ctrl+R",
        QtWidgets.QStyle.SP_DialogYesButton,
    )
    help_menu = menu.addMenu("&Help")
    _add_action("&About", gui.about, help_menu)
    _add_action("&Check for updates", gui.check_for_updates, help_menu)
    toolbar.setContextMenuPolicy(Qt.PreventContextMenu)
    return toolbar



from __future__ import annotations

import math

import numpy as np

import vstt
from vstt.experiment import Experiment


def test_data_df(experiment_with_results: Experiment) -> None:
    # results from 3 reps of 8-target trial, 1 rep of 4-target trial with automove back to center
    df = vstt.stats._data_df(experiment_with_results.trial_handler_with_results)
    assert len(df.columns) == len(vstt.stats._get_trial_data_columns())
    for _index, row in df.iterrows():
        assert row["to_target_mouse_positions"].shape[1] == 2
        assert (
            row["to_center_mouse_positions"].shape[0] == 0
            or row["to_center_mouse_positions"].shape[1] == 2
        )
        assert (
            row["to_target_timestamps"].shape[0]
            == row["to_target_mouse_positions"].shape[0]
        )
        assert (
            row["to_center_timestamps"].shape[0]
            == row["to_center_mouse_positions"].shape[0]
        )
    weights = [3, 1]
    targets = [8, 4]
    rows, cols = df.shape
    assert rows == np.sum(np.multiply(weights, targets))
    assert np.allclose(df["i_trial"].unique(), list(range(np.sum(weights))))
    assert df["i_rep"].unique() == [0]
    for i in [0, 1]:
        assert np.allclose(
            df.loc[df.condition_index == i]["i_target"].unique(),
            list(range(targets[i])),
        )


def test_stats_df(experiment_with_results: Experiment) -> None:
    df = vstt.stats.stats_dataframe(experiment_with_results.trial_handler_with_results)
    assert np.all(np.isnan(df.loc[df.condition_index == 1]["to_center_time"]))
    for _destination, stat_label_units in vstt.stats.list_dest_stat_label_units():
        for stat, _label, _unit in stat_label_units:
            assert stat in df.columns


def test_distance() -> None:
    assert np.allclose(vstt.stats._distance(np.array([[0, 0]])), [0])
    assert np.allclose(vstt.stats._distance(np.array([[3, 4]])), [0])
    assert np.allclose(vstt.stats._distance(np.array([[0, 0], [1, 1]])), [np.sqrt(2)])
    assert np.allclose(vstt.stats._distance(np.array([[1, 1], [0, 0]])), [np.sqrt(2)])
    assert np.allclose(
        vstt.stats._distance(np.array([[0, 0], [1, 1], [0, 0]])), [2.0 * np.sqrt(2)]
    )
    assert np.allclose(
        vstt.stats._distance(np.array([[0, 0], [1, 1], [0, 0], [-1, -1]])),
        [3.0 * np.sqrt(2)],
    )
    assert np.allclose(
        vstt.stats._distance(np.array([[0, 0], [1, 1], [0, 0], [-1, -1], [1, 1]])),
        [5.0 * np.sqrt(2)],
    )
    assert np.allclose(
        vstt.stats._distance(np.array([[0, 0], [np.sqrt(2), np.sqrt(2)], [0, 0]])), [4]
    )


def test_reaction_movement_time() -> None:
    for times in [
        [0.0, 0.1],
        [-1.0, -0.5],
        [0.0, 2.0, 4.0, 6.0],
        [-0.9, -0.7, -0.5, -0.3, -0.1, 0.1, 0.3, 0.5],
        [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.7, 1.0, 1.1, 1.2],
    ]:
        n = len(times)
        for n_zeros in range(1, n):
            for n_before_visible in range(1, n):
                positions = [[-1e-13, 1e-14]] * n_zeros + [[1, 1]] * (n - n_zeros)
                reaction_time = times[n_zeros] - times[n_before_visible]
                assert np.allclose(
                    vstt.stats._reaction_time(
                        np.array(times), np.array(positions), n_before_visible
                    ),
                    [reaction_time],
                )


def test_rmse() -> None:
    target = np.array([1, 1])
    # all points lie on the straight line between start and end point
    assert np.allclose(vstt.stats._rmse(np.array([(0, 0), (1, 1)]), target), [0])
    assert np.allclose(vstt.stats._rmse(np.array([(0, 0), (0.2, 0.2)]), target), [0])
    assert np.allclose(
        vstt.stats._rmse(
            np.array([(0, 0), (0.25, 0.25), (0.5, 0.5), (0.75, 0.75)]), target
        ),
        [0],
    )
    # points which are all 1/sqrt(2) perpendicular distance from line
    assert np.allclose(
        vstt.stats._rmse(np.array([(0, 0), (0, 1)]), target), [1.0 / np.sqrt(2.0)]
    )
    assert np.allclose(
        vstt.stats._rmse(np.array([(0, 0), (1, 0)]), target), [1.0 / np.sqrt(2.0)]
    )
    assert np.allclose(
        vstt.stats._rmse(np.array([(0, 0), (0, 1), (1, 0)]), target),
        [1.0 / np.sqrt(2.0)],
    )
    assert np.allclose(
        vstt.stats._rmse(np.array([(0, 0), (0, 1), (1, 0), (0, 1), (1, 0)]), target),
        [1.0 / np.sqrt(2.0)],
    )


def test_area() -> None:
    assert np.allclose(vstt.stats._area(np.array([]), np.array([])), [0])
    assert np.allclose(vstt.stats._area(np.array([]), np.array([[1, 1], [0, 1]])), [0])
    assert np.allclose(vstt.stats._area(np.array([[0, 0], [0, 1]]), np.array([])), [0])
    assert np.allclose(
        vstt.stats._area(
            np.array([[0, 0], [0, 1], [1, 1]]), np.array([[1, 1], [1, 0], [0, 0]])
        ),
        [1],
    )
    assert np.allclose(
        vstt.stats._area(np.array([[0, 0], [0, 1], [1, 1]]), np.array([])), [0.5]
    )
    assert np.allclose(
        vstt.stats._area(np.array([]), np.array([[0, 0], [0, 1], [1, 1]])), [0.5]
    )
    assert np.allclose(
        vstt.stats._area(
            np.array([[0, 1], [0, 0], [1, 1], [1, 0]]), np.array([[1, 0], [0, 1]])
        ),
        [0.5],
    )


def test_normalized_area() -> None:
    assert np.allclose(vstt.stats._normalized_area(np.array([]), np.array([])), [0])
    assert np.allclose(
        vstt.stats._normalized_area(np.array([]), np.array([[1, 1], [0, 1]])), [0]
    )
    assert np.allclose(
        vstt.stats._normalized_area(np.array([[0, 0], [0, 1]]), np.array([])), [0]
    )
    assert np.allclose(
        vstt.stats._normalized_area(
            np.array([[0, 0], [0, 1], [1, 1]]), np.array([[1, 1], [1, 0], [0, 0]])
        ),
        [1 / 16],
    )
    assert np.allclose(
        vstt.stats._normalized_area(np.array([[0, 0], [0, 1], [1, 1]]), np.array([])),
        [1 / (12 + 8 * math.sqrt(2))],
    )
    assert np.allclose(
        vstt.stats._normalized_area(np.array([]), np.array([[0, 0], [0, 1], [1, 1]])),
        [1 / (12 + 8 * math.sqrt(2))],
    )
    assert np.allclose(
        vstt.stats._normalized_area(
            np.array([[0, 1], [0, 0], [1, 1], [1, 0]]), np.array([[1, 0], [0, 1]])
        ),
        [1 / (24 + 16 * math.sqrt(2))],
    )


def test_peak_velocity() -> None:
    assert np.allclose(vstt.stats._peak_velocity(np.array([]), np.array([]))[0], [0])
    assert np.allclose(
        vstt.stats._peak_velocity(np.array([0, 0.5, 0.6, 1]), np.array([]))[0], [0]
    )
    assert np.allclose(
        vstt.stats._peak_velocity(
            np.array([]), np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        )[0],
        [0],
    )
    assert np.allclose(
        vstt.stats._peak_velocity(
            np.array([0, 0.5, 0.6, 1]), np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        )[0],
        [10],
    )
    assert np.allclose(
        vstt.stats._peak_velocity(np.array([0.5]), np.array([[0, 0]]))[0],
        [0],
    )
    assert np.allclose(
        vstt.stats._peak_velocity(np.array([0.5]), np.array([]))[0],
        [0],
    )
    assert np.allclose(
        vstt.stats._peak_velocity(np.array([]), np.array([[0, 0]]))[0],
        [0],
    )


def test_peak_acceleration() -> None:
    assert np.allclose(vstt.stats._peak_acceleration(np.array([]), np.array([])), [0])
    assert np.allclose(
        vstt.stats._peak_acceleration(np.array([0, 0.5, 0.6, 1]), np.array([])), [0]
    )
    assert np.allclose(
        vstt.stats._peak_acceleration(
            np.array([]), np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        ),
        [0],
    )
    assert np.allclose(
        vstt.stats._peak_acceleration(
            np.array([0, 0.5, 0.6, 1]), np.array([[0, 0], [0, 1], [1, 1], [1, 0]])
        ),
        [103.07764],
    )
    assert np.allclose(
        vstt.stats._peak_acceleration(np.array([0.5]), np.array([[0, 0]])),
        [0],
    )
    assert np.allclose(
        vstt.stats._peak_acceleration(np.array([0.5]), np.array([])),
        [0],
    )
    assert np.allclose(
        vstt.stats._peak_acceleration(np.array([]), np.array([[0, 0]])),
        [0],
    )


def test_spatial_error() -> None:
    assert np.allclose(
        vstt.stats._spatial_error(np.array([]), np.array([1, 1]), 0.01), [0]
    )
    assert np.allclose(
        vstt.stats._spatial_error(np.array([[1, 1]]), np.array([1, 1]), 0.01), [0]
    )
    assert np.allclose(
        vstt.stats._spatial_error(np.array([[0.99, 1]]), np.array([1, 1]), 0.02), [0]
    )
    assert np.allclose(
        vstt.stats._spatial_error(np.array([[-1, -1], [0, 0]]), np.array([1, 1]), 0.01),
        [1.404213562],
    )


def test_movement_time_at_peak_velocity() -> None:
    assert np.allclose(
        vstt.stats._movement_time_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4]]
            ),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._movement_time_at_peak_velocity(
            np.array([]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._movement_time_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]), np.array([]), 0
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._movement_time_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            0,
        ),
        0.02,
    )


def test_total_time_at_peak_velocity() -> None:
    assert np.allclose(
        vstt.stats._total_time_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4]]
            ),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._total_time_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            5,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._total_time_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            0,
        ),
        0.03,
    )


def test_movement_distance_at_peak_velocity() -> None:
    assert np.allclose(
        vstt.stats._movement_distance_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4]]
            ),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._movement_distance_at_peak_velocity(
            np.array([]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._movement_distance_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]), np.array([]), 0
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._movement_distance_at_peak_velocity(
            np.array([0.09, 0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [
                    [0, 0],
                    [0, 0],
                    [0.1, 0.4],
                    [0.4, 0.6],
                    [0.6, 0.75],
                    [1, 1.2],
                    [1.1, 1.5],
                ]
            ),
            0,
        ),
        0.611,
        rtol=1e-3,
    )


def test_rmse_movement_at_peak_velocity() -> None:
    assert np.allclose(
        vstt.stats._rmse_movement_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4], [0.1, 0.4]]
            ),
            np.array([1.1, 1.5]),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._rmse_movement_at_peak_velocity(
            np.array([]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            np.array([1.1, 1.5]),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._rmse_movement_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array([]),
            np.array([1.1, 1.5]),
            0,
        ),
        np.nan,
        equal_nan=True,
    )

    assert np.allclose(
        vstt.stats._rmse_movement_at_peak_velocity(
            np.array([0.1, 0.11, 0.12, 0.13, 0.14, 0.15]),
            np.array(
                [[0, 0], [0.1, 0.4], [0.4, 0.6], [0.6, 0.75], [1, 1.2], [1.1, 1.5]]
            ),
            np.array([1.1, 1.5]),
            0,
        ),
        0.13453,
        rtol=1e-3,
    )



from __future__ import annotations

import qt_test_utils as qtu
from psychopy.visual.window import Window

import vstt
from vstt.display_widget import DisplayOptionsWidget
from vstt.experiment import Experiment


def test_display_options_widget(window: Window) -> None:
    widget = DisplayOptionsWidget(parent=None, win=window)
    signal_received = qtu.SignalReceived(widget.experiment_modified)
    # initially has default experiment with default display options
    assert widget.experiment.display_options == vstt.display.default_display_options()
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    # set an experiment with all options set to false
    experiment = Experiment()
    all_false = vstt.display.default_display_options()
    for key in all_false:
        all_false[key] = False  # type: ignore
    experiment.display_options = all_false
    widget.experiment = experiment
    assert widget.experiment is experiment
    assert widget.experiment.display_options == all_false
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    # modify options by clicking on each check box in turn
    for key, check_box in widget._widgets.items():
        signal_received.clear()
        assert widget.experiment.display_options[key] is False  # type: ignore
        qtu.click(check_box)
        assert widget.experiment.display_options[key] is True  # type: ignore
        assert widget.experiment.has_unsaved_changes is True
        assert signal_received
        signal_received.clear()
        qtu.click(check_box)
        assert widget.experiment.display_options[key] is False  # type: ignore
        assert signal_received
    # check that all values have the correct type
    for value in widget.experiment.display_options.values():
        assert isinstance(value, bool)



from __future__ import annotations

import threading
from copy import deepcopy
from time import sleep

import gui_test_utils as gtu
import numpy as np
import pyautogui
import pytest
from psychopy.visual.window import Window

import vstt
from vstt.experiment import Experiment
from vstt.geom import points_on_circle
from vstt.task import MotorTask


def move_mouse_to_target(
    target_pixel: tuple[float, float],
    target_color: tuple[int, int, int] = (255, 0, 0),
    movement_time: float = 0.1,
) -> None:
    # wait until target is activated
    attempts = 0
    sleep_interval = 0.1
    while not gtu.pixel_color(target_pixel, target_color):
        sleep(sleep_interval)
        sleep_interval += 0.1
        # press Enter in case we are looking at a splash / display_results screen
        gtu.press_enter()
        attempts += 1
        if attempts > 10:
            print(f"Stuck waiting for target at {target_pixel} to activate", flush=True)
            gtu.ascii_screenshot()
    gtu.ascii_screenshot()
    # move mouse to target
    pyautogui.moveTo(target_pixel[0], target_pixel[1], movement_time)


def do_task(
    experiment: Experiment, target_pixels: list[list[tuple[float, float]]]
) -> None:
    gtu.ascii_screenshot()
    for target_pixels_block, trial in zip(target_pixels, experiment.trial_list):
        for _rep in range(trial["weight"]):
            for target_pixel in target_pixels_block:
                move_mouse_to_target(target_pixel)


def launch_do_task(
    experiment: Experiment, target_pixels: list[list[tuple[float, float]]]
) -> threading.Thread:
    thread = threading.Thread(
        target=do_task,
        name="task",
        args=(
            experiment,
            target_pixels,
        ),
    )
    thread.start()
    return thread


def test_task_no_trials(window: Window) -> None:
    experiment_no_trials = Experiment()
    experiment_no_trials.trial_list = []
    experiment_no_trials.has_unsaved_changes = False
    assert experiment_no_trials.trial_handler_with_results is None
    do_task_thread = launch_do_task(experiment_no_trials, [])
    task = MotorTask(experiment_no_trials, window)
    success = task.run()
    do_task_thread.join()
    assert success is False
    assert experiment_no_trials.has_unsaved_changes is False
    assert experiment_no_trials.trial_handler_with_results is None


@pytest.mark.parametrize(
    "trial_settings",
    [
        {"automove_cursor_to_center": True},
        {"add_central_target": False, "automove_cursor_to_center": False, "weight": 1},
    ],
    ids=["automove_to_center", "no_central_target"],
)
def test_task(
    experiment_no_results: Experiment, window: Window, trial_settings: dict
) -> None:
    target_pixels = []
    experiment_no_results.has_unsaved_changes = False
    assert experiment_no_results.trial_handler_with_results is None
    for trial in experiment_no_results.trial_list:
        for key, value in trial_settings.items():
            trial[key] = value  # type: ignore
        target_pixels.append(
            [
                gtu.pos_to_pixels(pos)
                for pos in points_on_circle(
                    trial["num_targets"],
                    trial["target_distance"],
                    include_centre=False,
                )
            ]
        )
    do_task_thread = launch_do_task(experiment_no_results, target_pixels)
    task = MotorTask(experiment_no_results, window)
    success = task.run()
    do_task_thread.join()
    # task ran successfully, updated experiment with results
    assert success is True
    assert experiment_no_results.has_unsaved_changes is True
    assert experiment_no_results.trial_handler_with_results is not None
    data = experiment_no_results.trial_handler_with_results.data
    # check that we hit all the targets without timing out
    for timestamps in data["to_target_timestamps"][0][0]:
        assert (
            timestamps[-1] - timestamps[0]
            < 0.8 * experiment_no_results.trial_list[0]["target_duration"]
        )
    for dest in ["target", "center"]:
        assert np.all(data[f"to_{dest}_success"][0][0])


def test_task_no_automove_to_center(
    experiment_no_results: Experiment, window: Window
) -> None:
    experiment_no_results.has_unsaved_changes = False
    assert experiment_no_results.trial_handler_with_results is None
    target_pixels = []
    for trial in experiment_no_results.trial_list:
        trial["automove_cursor_to_center"] = False
        tp = []
        for pos in points_on_circle(
            trial["num_targets"],
            trial["target_distance"],
            include_centre=False,
        ):
            # after moving to each target need to return mouse to central target
            tp.append(gtu.pos_to_pixels(pos))
            tp.append(gtu.pos_to_pixels((0, 0)))
        target_pixels.append(tp)
    do_task_thread = launch_do_task(experiment_no_results, target_pixels)
    task = vstt.task.MotorTask(experiment_no_results, window)
    success = task.run()
    do_task_thread.join()
    # task ran successfully, updated experiment with results
    assert success is True
    assert experiment_no_results.has_unsaved_changes is True
    assert experiment_no_results.trial_handler_with_results is not None
    # check that we hit all the targets
    data = experiment_no_results.trial_handler_with_results.data
    for to_target_timestamps, to_center_timestamps in zip(
        data["to_target_timestamps"][0][0],
        data["to_center_timestamps"][0][0],
    ):
        assert (
            to_target_timestamps[-1] - to_target_timestamps[0]
            < 0.8 * experiment_no_results.trial_list[0]["target_duration"]
        )
        assert (
            to_center_timestamps[-1] - to_target_timestamps[0]
            < 0.8 * experiment_no_results.trial_list[0]["target_duration"]
        )
    for dest in ["target", "center"]:
        assert np.all(data[f"to_{dest}_success"][0][0])


def test_task_fixed_intervals_no_user_input(window: Window) -> None:
    experiment = Experiment()
    experiment.metadata["display_duration"] = 0.0
    target_duration = 1.0
    trial = vstt.trial.default_trial()
    trial["weight"] = 3
    trial["num_targets"] = 3
    trial["target_order"] = "random"
    trial["show_target_labels"] = True
    trial["fixed_target_intervals"] = True
    trial["target_duration"] = target_duration
    trial["automove_cursor_to_center"] = False
    trial["freeze_cursor_between_targets"] = False
    trial["post_block_delay"] = 0.0
    trial["play_sound"] = False
    experiment.trial_list = [trial]
    assert experiment.trial_handler_with_results is None
    task = vstt.task.MotorTask(experiment, window)
    # run task without moving mouse, which will stay in center for entire experiment
    assert task.run() is True
    # task ran successfully, updated experiment with results
    assert experiment.has_unsaved_changes is True
    assert experiment.trial_handler_with_results is not None
    # check that we failed to hit all targets
    expected_success = np.full((trial["num_targets"],), False)
    data = experiment.trial_handler_with_results.data
    assert data["to_target_timestamps"].shape == (3, 1)
    for irep in [0, 1, 2]:
        for success_name in ["to_target_success", "to_center_success"]:
            assert np.all(data[success_name][irep][0] == expected_success)
        # For first target of first repetition: to_target timestamps should go from 0 to 2*target_duration
        # All subsequent to_target timestamps should follow sequentially and last target_duration
        # Each rep resets the clock to zero
        all_to_target_timestamps = data["to_target_timestamps"][irep][0]
        # require timestamps to be accurate within 0.5s
        allowed_error_on_timestamp = 0.5  # 1.0/30.0
        # this is a weak requirement to avoid tests failing on CI where many frames can get dropped
        # if running tests locally allowed_error_on_timestamp should be ~2/fps
        expected_initial_t_first_target = 0
        expected_final_t_first_target = (
            2 * target_duration if irep == 0 else target_duration
        )
        assert (
            abs(all_to_target_timestamps[0][0]) - expected_initial_t_first_target
            < allowed_error_on_timestamp
        )
        assert (
            abs(all_to_target_timestamps[0][-1] - expected_final_t_first_target)
            < allowed_error_on_timestamp
        )
        for count, to_target_timestamps in enumerate(all_to_target_timestamps[1:]):
            expected_initial_t = expected_final_t_first_target + count * target_duration
            expected_final_t = expected_initial_t + target_duration
            assert (
                abs(to_target_timestamps[0] - expected_initial_t)
                < allowed_error_on_timestamp
            )
            assert (
                abs(to_target_timestamps[-1] - expected_final_t)
                < allowed_error_on_timestamp
            )


def test_task_condition_timeout_no_user_input(window: Window) -> None:
    # 4 seconds condition timeout where each trial has 1 target which is displayed for up to 1.3 seconds
    # first target of condition has additional 0.20 seconds delay
    # should get 2 trials in ~2.80secs, 3rd trial should time out
    experiment = Experiment()
    experiment.metadata["display_duration"] = 0.0
    trial1 = vstt.trial.default_trial()
    trial1["weight"] = 100
    trial1["condition_timeout"] = 4.0
    trial1["num_targets"] = 1
    trial1["target_duration"] = 1.30
    trial1["pre_target_delay"] = 0.0
    trial1["pre_central_target_delay"] = 0.0
    trial1["pre_first_target_extra_delay"] = 0.20
    trial1["add_central_target"] = False
    trial1["automove_cursor_to_center"] = False
    trial1["freeze_cursor_between_targets"] = False
    trial1["post_block_delay"] = 0.0
    trial1["play_sound"] = False
    # 6 seconds timeout: 3 targets, each for 1.05 seconds -> single completed trial
    trial2 = deepcopy(trial1)
    trial2["condition_timeout"] = 4.0
    trial2["num_targets"] = 3
    trial2["target_duration"] = 0.7
    experiment.trial_list = [trial1, trial2]
    assert experiment.trial_handler_with_results is None
    task = vstt.task.MotorTask(experiment, window)
    # run task without moving mouse, which will stay in center for entire experiment
    assert task.run() is True
    # task ran successfully, updated experiment with results
    assert experiment.has_unsaved_changes is True
    assert experiment.trial_handler_with_results is not None
    experiment.stats.to_pickle("./df.pkl")
    # we should get exactly 3 trials
    assert len(experiment.stats.i_trial.unique()) == 3
    # first two trials have 1 target, last trial has 3 targets -> 5 targets total
    assert len(experiment.stats.to_target_success) == 5
    # all targets should have been missed
    assert np.all(~experiment.stats.to_target_success)



from __future__ import annotations

from typing import Callable

from psychopy.visual.window import Window
from qtpy import QtCore
from qtpy import QtWidgets

from vstt.display import display_options_labels
from vstt.experiment import Experiment


class DisplayOptionsWidget(QtWidgets.QWidget):
    experiment_modified = QtCore.Signal()

    def __init__(
        self, parent: QtWidgets.QWidget | None = None, win: Window | None = None
    ):
        super().__init__(parent)
        self._win = win
        self._experiment: Experiment = Experiment()
        self._widgets: dict[str, QtWidgets.QCheckBox] = {}

        outer_layout = QtWidgets.QVBoxLayout()
        group_box = QtWidgets.QGroupBox("Display Options")
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(group_box)
        outer_layout.addWidget(scroll_area)
        inner_layout = QtWidgets.QVBoxLayout()
        group_box.setLayout(inner_layout)
        labels = display_options_labels()
        for _row_index, key in enumerate(labels.keys()):
            checkbox = QtWidgets.QCheckBox(f"{labels[key]}", self)
            inner_layout.addWidget(checkbox)
            checkbox.clicked.connect(self._update_value_callback(key))
            self._widgets[key] = checkbox
        self.setLayout(outer_layout)

    def _update_value_callback(self, key: str) -> Callable[[bool], None]:
        def _update_value(value: bool) -> None:
            self._experiment.display_options[key] = value  # type: ignore
            self._experiment.has_unsaved_changes = True
            self.experiment_modified.emit()

        return _update_value

    @property
    def experiment(self) -> Experiment:
        return self._experiment

    @experiment.setter
    def experiment(self, experiment: Experiment) -> None:
        self._experiment = experiment
        for key, widget in self._widgets.items():
            widget.setChecked(self._experiment.display_options[key])  # type: ignore



from __future__ import annotations

from typing import Callable

from psychopy.visual.window import Window
from qtpy import QtCore
from qtpy import QtWidgets
from qtpy.QtCore import Qt

from vstt.experiment import Experiment
from vstt.meta import default_metadata
from vstt.meta import metadata_labels
from vstt.vis import splash_screen


class MetadataWidget(QtWidgets.QWidget):
    experiment_modified = QtCore.Signal()

    def __init__(
        self, parent: QtWidgets.QWidget | None = None, win: Window | None = None
    ):
        super().__init__(parent)
        self._win = win
        self._experiment = Experiment()
        self._str_widgets: dict[str, QtWidgets.QLineEdit] = {}
        self._bool_widgets: dict[str, QtWidgets.QCheckBox] = {}
        self._float_widgets: dict[str, QtWidgets.QDoubleSpinBox] = {}

        outer_layout = QtWidgets.QVBoxLayout()
        group_box = QtWidgets.QGroupBox("Metadata")
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(group_box)
        outer_layout.addWidget(scroll_area)
        inner_layout = QtWidgets.QVBoxLayout()
        group_box.setLayout(inner_layout)
        fields_layout = QtWidgets.QGridLayout()
        fields = QtWidgets.QWidget()
        fields.setLayout(fields_layout)
        labels = metadata_labels()
        for row_index, (key, value) in enumerate(default_metadata().items()):
            if isinstance(value, str):
                lbl = QtWidgets.QLabel(f"{labels[key]}:", self)
                lbl.setAlignment(Qt.AlignRight)
                fields_layout.addWidget(lbl, row_index, 0)
                line_edit = QtWidgets.QLineEdit(self)
                fields_layout.addWidget(line_edit, row_index, 1)
                line_edit.textEdited.connect(self._update_value_callback(key))
                self._str_widgets[key] = line_edit
            elif isinstance(value, bool):
                checkbox = QtWidgets.QCheckBox(f"{labels[key]}", self)
                fields_layout.addWidget(checkbox, row_index, 1)
                checkbox.clicked.connect(self._update_value_callback(key))
                self._bool_widgets[key] = checkbox
            elif isinstance(value, (int, float)):
                lbl = QtWidgets.QLabel(f"{labels[key]}:", self)
                lbl.setAlignment(Qt.AlignRight)
                fields_layout.addWidget(lbl, row_index, 0)
                spinbox = QtWidgets.QDoubleSpinBox(self)
                fields_layout.addWidget(spinbox, row_index, 1)
                spinbox.valueChanged.connect(self._update_value_callback(key))
                self._float_widgets[key] = spinbox

        inner_layout.addWidget(fields)
        self._btn_preview_metadata = QtWidgets.QPushButton("Preview Splash Screen")
        self._btn_preview_metadata.clicked.connect(self._btn_preview_metadata_clicked)
        inner_layout.addWidget(self._btn_preview_metadata)
        self.setLayout(outer_layout)

    def _update_value_callback(self, key: str) -> Callable[[str | bool | float], None]:
        def _update_value(value: str | bool | float) -> None:
            if value != self._experiment.metadata[key]:  # type: ignore
                self._experiment.metadata[key] = value  # type: ignore
                self._experiment.has_unsaved_changes = True
                self.experiment_modified.emit()

        return _update_value

    def _btn_preview_metadata_clicked(self) -> None:
        splash_screen(
            display_time_seconds=self._experiment.metadata["display_duration"],
            enter_to_skip_delay=self._experiment.metadata["enter_to_skip_delay"],
            show_delay_countdown=self._experiment.metadata["show_delay_countdown"],
            metadata=self._experiment.metadata,
            win=self._win,
        )

    @property
    def experiment(self) -> Experiment:
        return self._experiment

    @experiment.setter
    def experiment(self, experiment: Experiment) -> None:
        self._experiment = experiment
        for key, str_widget in self._str_widgets.items():
            str_widget.setText(self._experiment.metadata[key])  # type: ignore
        for key, bool_widget in self._bool_widgets.items():
            bool_widget.setChecked(self._experiment.metadata[key])  # type: ignore
        for key, float_widget in self._float_widgets.items():
            float_widget.setValue(self._experiment.metadata[key])  # type: ignore



from __future__ import annotations

import sys

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class Trial(TypedDict):
    weight: int
    condition_timeout: float
    num_targets: int
    target_order: str | list
    target_indices: str
    add_central_target: bool
    hide_target_when_reached: bool
    turn_target_to_green_when_reached: bool
    show_target_labels: bool
    target_labels: str
    fixed_target_intervals: bool
    target_duration: float
    central_target_duration: float
    pre_target_delay: float
    pre_central_target_delay: float
    pre_first_target_extra_delay: float
    target_distance: float
    target_size: float
    central_target_size: float
    show_inactive_targets: bool
    ignore_incorrect_targets: bool
    play_sound: bool
    use_joystick: bool
    joystick_max_speed: float
    show_cursor: bool
    cursor_size: float
    show_cursor_path: bool
    automove_cursor_to_center: bool
    freeze_cursor_between_targets: bool
    cursor_rotation_degrees: float
    post_trial_delay: float
    post_trial_display_results: bool
    post_block_delay: float
    post_block_display_results: bool
    show_delay_countdown: bool
    enter_to_skip_delay: bool


class DisplayOptions(TypedDict):
    to_target_paths: bool
    to_center_paths: bool
    targets: bool
    central_target: bool
    to_target_reaction_time: bool
    to_center_reaction_time: bool
    to_target_movement_time: bool
    to_center_movement_time: bool
    to_target_time: bool
    to_center_time: bool
    to_target_distance: bool
    to_center_distance: bool
    to_target_rmse: bool
    to_center_rmse: bool
    to_target_success: bool
    to_center_success: bool
    averages: bool
    area: bool
    normalized_area: bool
    peak_velocity: bool
    peak_acceleration: bool
    to_target_spatial_error: bool
    to_center_spatial_error: bool
    movement_time_at_peak_velocity: bool
    total_time_at_peak_velocity: bool
    movement_distance_at_peak_velocity: bool
    rmse_movement_at_peak_velocity: bool


class Metadata(TypedDict):
    name: str
    subject: str
    date: str
    author: str
    display_title: str
    display_text1: str
    display_text2: str
    display_text3: str
    display_text4: str
    display_duration: float
    show_delay_countdown: bool
    enter_to_skip_delay: bool



from __future__ import annotations

import numpy as np
import pandas as pd
from PIL.Image import Image
from psychopy.clock import Clock
from psychopy.colors import colorNames
from psychopy.data import TrialHandlerExt
from psychopy.event import Mouse
from psychopy.hardware.keyboard import Keyboard
from psychopy.visual.basevisual import BaseVisualStim
from psychopy.visual.circle import Circle
from psychopy.visual.elementarray import ElementArrayStim
from psychopy.visual.shape import ShapeStim
from psychopy.visual.textbox2 import TextBox2
from psychopy.visual.window import Window

from vstt.geom import points_on_circle
from vstt.stats import get_closed_polygon
from vstt.stats import list_dest_stat_label_units
from vstt.stats import stats_dataframe
from vstt.vtypes import DisplayOptions
from vstt.vtypes import Metadata

colors = [
    color for name, color in colorNames.items() if name not in ["none", "transparent"]
]


def make_cursor(window: Window, cursor_size: float) -> ShapeStim:
    return ShapeStim(
        window,
        lineColor="black",
        pos=(0, 0),
        lineWidth=5,
        vertices=[
            (-cursor_size * 0.5, 0.00),
            (cursor_size * 0.5, 0.00),
            (0.00, 0.00),
            (0.00, -cursor_size * 0.5),
            (0.00, cursor_size * 0.5),
        ],
        closeShape=False,
    )


def make_targets(
    window: Window,
    n_circles: int,
    radius: float,
    point_radius: float,
    add_central_target: bool,
    center_point_radius: float,
) -> ElementArrayStim:
    n_elements = n_circles + 1 if add_central_target else n_circles
    sizes = [[2.0 * point_radius] * 2] * n_circles
    if add_central_target:
        sizes = sizes + [[2.0 * center_point_radius] * 2]
    return ElementArrayStim(
        window,
        units="height",
        fieldShape="circle",
        nElements=n_elements,
        sizes=sizes,
        xys=points_on_circle(n_circles, radius, include_centre=add_central_target),
        elementTex=None,
        elementMask="circle",
    )


def make_target_labels(
    window: Window,
    n_circles: int,
    radius: float,
    point_radius: float,
    labels_string: str,
) -> list[TextBox2]:
    text_boxes: list[TextBox2] = []
    positions = points_on_circle(n_circles, radius, include_centre=False)
    labels = labels_string.strip().split(" ")
    for label, position in zip(
        labels,
        positions,
    ):
        text_boxes.append(
            TextBox2(
                window,
                label,
                anchor="center",
                pos=position,
                color="white",
                bold=True,
                alignment="center",
                letterHeight=1.25 * point_radius,
            )
        )
    return text_boxes


def update_target_colors(
    targets: ElementArrayStim,
    show_inactive_targets: bool,
    index: int | None = None,
    green_target_index: int | None = None,
) -> None:
    inactive_rgb = 0.0
    if show_inactive_targets:
        inactive_rgb = 0.9
    c = np.array([[inactive_rgb, inactive_rgb, inactive_rgb]] * targets.nElements)
    if index is not None:
        # Set specified target to red
        c[index] = [1, -1, -1]
    if green_target_index is not None:
        # Set specified target to green
        c[green_target_index] = [-1, 1, -1]
    targets.setColors(c, colorSpace="rgb")


def update_target_label_colors(
    target_labels: list[TextBox2],
    show_inactive_targets: bool,
    index: int | None = None,
) -> None:
    inactive_rgb = 0.0
    if show_inactive_targets:
        inactive_rgb = 0.3
    for target_index, target_label in enumerate(target_labels):
        if target_index == index:
            target_label.color = (1, 1, 1)
        else:
            target_label.color = (inactive_rgb, inactive_rgb, inactive_rgb)


class MotorTaskCancelledByUser(Exception):
    pass


def draw_and_flip(
    win: Window,
    drawables: list[BaseVisualStim],
    kb: Keyboard | None,
    kb_stop_key: str = "escape",
) -> None:
    for drawable in drawables:
        drawable.draw()
    win.flip()
    if kb is not None and kb.getKeys([kb_stop_key]):
        raise MotorTaskCancelledByUser


def _make_stats_txt(
    display_options: DisplayOptions,
    stats: pd.Series,
    all_trials_for_this_condition: bool,
    average: bool,
) -> str:
    txt_stats = ""
    for destination, stat_label_units in list_dest_stat_label_units():
        for stat, label, unit in stat_label_units:
            if display_options.get(stat, False):  # type: ignore
                stat_str = f"{stats[stat]: .3f}{unit}"
                if stat == f"to_{destination}_success":
                    if average or all_trials_for_this_condition:
                        stat_str = f"{stats[stat]: .0%}"
                    else:
                        stat_str = f"{stats[stat] == 1}"
                if stat in [
                    "area",
                    "normalized_area",
                    "peak_velocity",
                    "peak_acceleration",
                    "movement_time_at_peak_velocity",
                    "total_time_at_peak_velocity",
                    "movement_distance_at_peak_velocity",
                    "rmse_movement_at_peak_velocity",
                ]:
                    txt_stats += f"{label}: {stat_str}\n"
                else:
                    txt_stats += f"{label} (to {destination}): {stat_str}\n"
    return txt_stats


def _make_stats_drawables(
    trial_handler: TrialHandlerExt,
    display_options: DisplayOptions,
    stats_df: pd.DataFrame,
    win: Window,
    all_trials_for_this_condition: bool,
) -> list[BaseVisualStim]:
    drawables: list[BaseVisualStim] = []
    if stats_df.shape[0] == 0:
        # no results to display
        return drawables
    i_condition = stats_df.iloc[0].condition_index
    conditions = trial_handler.trialList[i_condition]
    trial_indices = stats_df.i_trial.unique()
    drawables.append(
        TextBox2(
            win,
            f"Condition {i_condition}, trial{'s' if len(trial_indices) > 1 else ''} {trial_indices}",
            anchor="top_center",
            pos=(0, 0.5),
            color="black",
            alignment="top_center",
            letterHeight=0.02,
        )
    )
    # stats
    letter_height = 0.014
    # split target_pos pair into two scalars, so they survive the groupby.mean()
    stats_df[["target_pos_x", "target_pos_y"]] = pd.DataFrame(
        stats_df.target_pos.tolist(), index=stats_df.index
    )
    # convert to_target_success and to_center_success into scalars 0/1
    stats_df.to_target_success = stats_df.to_target_success.astype(int)
    stats_df.to_center_success = stats_df.to_center_success.astype(int)

    for _, row in (
        stats_df.groupby("target_index", as_index=False)
        .mean(numeric_only=True)
        .iterrows()
    ):
        color = colors[int(np.rint(row.target_index))]
        txt_stats = _make_stats_txt(
            display_options, row, all_trials_for_this_condition, average=False
        )
        if row.target_pos_x > 0:
            text_pos = row.target_pos_x + 0.18, row.target_pos_y
        else:
            text_pos = row.target_pos_x - 0.18, row.target_pos_y
        n_lines = txt_stats.count("\n")
        if n_lines <= 4:
            letter_height = 0.028
        if n_lines <= 6:
            letter_height = 0.018
        if len(txt_stats) > 0:
            drawables.append(
                TextBox2(
                    win,
                    txt_stats,
                    pos=text_pos,
                    color=color,
                    alignment="center",
                    letterHeight=letter_height,
                )
            )

    # update the values of to_target_success and to_center_success to the successful trial fraction or successful target fraction
    for dest in ["target", "center"]:
        stats_df[f"to_{dest}_success"] = (
            get_successful_trial_fraction(stats_df, dest)
            if all_trials_for_this_condition
            else get_successful_target_fraction(stats_df, dest)
        )

    if display_options["averages"]:
        txt_stats = "Averages:\n" + _make_stats_txt(
            display_options,
            stats_df.mean(numeric_only=True),
            all_trials_for_this_condition,
            average=True,
        )
        drawables.append(
            TextBox2(
                win,
                txt_stats,
                anchor="top_left",
                pos=(-0.5 * win.size[0] / win.size[1], 0.5),
                color="black",
                alignment="top_left",
                letterHeight=letter_height,
            )
        )
    # central target
    if (
        display_options["central_target"]
        and not conditions["automove_cursor_to_center"]
    ):
        drawables.append(
            Circle(
                win,
                radius=conditions["central_target_size"],
                pos=(0.0, 0.0),
                fillColor=(0.1, 0.1, 0.1),
            )
        )
    # targets
    if display_options["targets"]:
        for _, row in stats_df.loc[stats_df.i_trial == trial_indices[0]].iterrows():
            drawables.append(
                Circle(
                    win,
                    radius=conditions["target_size"],
                    pos=row.target_pos,
                    fillColor=colors[row.target_index],
                )
            )
    # paths to center
    if (
        display_options["to_center_paths"]
        and conditions["add_central_target"]
        and not conditions["automove_cursor_to_center"]
    ):
        for _, row in stats_df.iterrows():
            drawables.append(
                ShapeStim(
                    win,
                    vertices=row.to_center_mouse_positions,
                    lineColor=colors[row.target_index],
                    closeShape=False,
                    lineWidth=3,
                )
            )
    # paths to target
    if display_options["to_target_paths"]:
        for _, row in stats_df.iterrows():
            drawables.append(
                ShapeStim(
                    win,
                    vertices=row.to_target_mouse_positions,
                    lineColor=colors[row.target_index],
                    closeShape=False,
                    lineWidth=3,
                )
            )
    # fill the area closed by paths to target and center
    if display_options["area"]:
        for _, row in stats_df.iterrows():
            drawables.append(
                ShapeStim(
                    win,
                    vertices=get_closed_polygon(
                        row.to_target_mouse_positions, row.to_center_mouse_positions
                    ),
                    lineColor="black",
                    fillColor=colors[row.target_index],
                    closeShape=True,
                    lineWidth=3,
                )
            )
    return drawables


def display_results(
    display_time_seconds: float,
    enter_to_skip_delay: bool,
    show_delay_countdown: bool,
    trial_handler: TrialHandlerExt | None,
    display_options: DisplayOptions,
    i_trial: int,
    all_trials_for_this_condition: bool,
    win: Window | None = None,
    mouse: Mouse | None = None,
    mouse_pos: tuple[float, float] | None = None,
    return_screenshot: bool = False,
) -> Image | None:
    close_window_when_done = False
    if win is None:
        win = _make_window()
        close_window_when_done = True
    drawables = []
    if trial_handler is not None:
        stats_df = stats_dataframe(trial_handler)
        if all_trials_for_this_condition:
            condition_index = next(
                iter(
                    stats_df.loc[stats_df.i_trial == i_trial].condition_index.to_numpy()
                ),
                -1,
            )
            stats_df = stats_df.loc[stats_df.condition_index == condition_index]
        else:
            stats_df = stats_df.loc[stats_df.i_trial == i_trial]
        drawables = _make_stats_drawables(
            trial_handler, display_options, stats_df, win, all_trials_for_this_condition
        )
    return display_drawables(
        display_time_seconds,
        enter_to_skip_delay,
        show_delay_countdown,
        drawables,
        win,
        close_window_when_done,
        mouse,
        mouse_pos,
        return_screenshot,
    )


def get_successful_trial_fraction(stats_df: pd.DataFrame, dest: str) -> float:
    """
    get successful trial fraction

    :param stats_df: stats dataframe
    :param dest: "target" or "center"
    :return: successful trial fraction

    """
    trial_indices = stats_df["i_trial"].unique()

    to_dest_success_trial = 0
    for trial_index in trial_indices:
        data = stats_df.loc[stats_df.i_trial == trial_index]
        to_dest_success_target_number = (data[f"to_{dest}_success"]).values.sum()
        to_dest_success_trial += 1 if to_dest_success_target_number == len(data) else 0
    successful_trial_fraction = to_dest_success_trial / len(trial_indices)

    return successful_trial_fraction


def get_successful_target_fraction(stats_df: pd.DataFrame, dest: str) -> float:
    """
    get successful target fraction

    :param stats_df: stats dataframe
    :param dest: "target" or "center"
    :return: successful target fraction

    """
    successful_target_fraction = stats_df[f"to_{dest}_success"].values.sum() / len(
        stats_df[f"to_{dest}_success"]
    )
    return successful_target_fraction


def _make_textbox_press_enter(win: Window) -> TextBox2:
    return TextBox2(
        win,
        "Please press Enter when you are ready to continue...",
        pos=(0, -0.47),
        color="navy",
        alignment="center",
        letterHeight=0.03,
    )


def _make_textbox_title(title: str, win: Window) -> TextBox2:
    return TextBox2(
        win,
        title,
        pos=(0, 0.40),
        size=(1, None),
        color="black",
        bold=True,
        alignment="center",
        letterHeight=0.06,
    )


def _make_textbox_main_text(text: str, win: Window) -> TextBox2:
    return TextBox2(
        win,
        text,
        size=(1, None),
        pos=(0, 0),
        color="black",
        alignment="left",
        letterHeight=0.03,
    )


def _make_textbox_countdown(text: str, win: Window) -> TextBox2:
    return TextBox2(
        win,
        text,
        size=(1, None),
        pos=(0, -0.3),
        color="black",
        alignment="center",
        letterHeight=0.2,
    )


def _make_window() -> Window:
    return Window(fullscr=True, units="height")


def display_drawables(
    display_time_seconds: float,
    enter_to_skip_delay: bool,
    show_delay_countdown: bool,
    drawables: list[BaseVisualStim],
    win: Window,
    close_window_when_done: bool,
    mouse: Mouse | None = None,
    mouse_pos: tuple[float, float] | None = None,
    return_screenshot: bool = False,
) -> Image | None:
    screenshot = None
    if drawables is None:
        drawables = []
    if return_screenshot:
        # do one flip, take screenshot, return image
        enter_to_skip_delay = False
        draw_and_flip(win, drawables, None)
        screenshot = win.getMovieFrame()
    else:
        remaining_display_time = int(np.ceil(display_time_seconds))
        if enter_to_skip_delay:
            kb = Keyboard()
            kb.clearEvents()
            drawables.append(_make_textbox_press_enter(win))
        else:
            kb = None
        if show_delay_countdown:
            countdown_textbox = _make_textbox_countdown(
                f"{remaining_display_time}", win
            )
            drawables.append(countdown_textbox)
        clock = Clock()
        clock.reset()
        while clock.getTime() < display_time_seconds:
            if show_delay_countdown:
                new_remaining_display_time = int(
                    np.ceil(display_time_seconds - clock.getTime())
                )
                if new_remaining_display_time < remaining_display_time:
                    remaining_display_time = new_remaining_display_time
                    countdown_textbox.text = f"{remaining_display_time}"
            try:
                if mouse is not None and mouse_pos is not None:
                    mouse.setPos(mouse_pos)
                draw_and_flip(win, drawables, kb, "return")
            except MotorTaskCancelledByUser:
                if close_window_when_done:
                    win.close()
                return None
    if close_window_when_done:
        win.close()
    return screenshot


def splash_screen(
    display_time_seconds: float,
    enter_to_skip_delay: bool,
    show_delay_countdown: bool,
    metadata: Metadata,
    win: Window | None = None,
) -> None:
    close_window_when_done = False
    if win is None:
        win = _make_window()
        close_window_when_done = True
    drawables = [_make_textbox_title(metadata["display_title"], win)]
    main_text = "\n\n".join(
        [
            metadata["display_text1"],
            metadata["display_text2"],
            metadata["display_text3"],
            metadata["display_text4"],
        ]
    )
    drawables.append(_make_textbox_main_text(main_text, win))
    display_drawables(
        display_time_seconds,
        enter_to_skip_delay,
        show_delay_countdown,
        drawables,
        win,
        close_window_when_done,
    )



from __future__ import annotations

import copy
import logging
from typing import Any
from typing import Mapping
from typing import TypeVar

from vstt.vtypes import DisplayOptions
from vstt.vtypes import Metadata
from vstt.vtypes import Trial

VsttTypedDict = TypeVar(
    "VsttTypedDict",
    Trial,
    Metadata,
    DisplayOptions,
)


def _has_valid_type(var: Any, correct_type: type) -> bool:
    if isinstance(var, correct_type):
        # var has the correct type
        return True
    # int instead of float or vice versa is also ok - anything else is not
    return correct_type in (float, int) and type(var) in (float, int)


def import_typed_dict(
    input_dict: Mapping[str, Any], default_typed_dict: VsttTypedDict
) -> VsttTypedDict:
    # start with a valid typed dict with default values
    output_dict = copy.deepcopy(default_typed_dict)
    for key, default_value in output_dict.items():
        if key in input_dict:
            # import all valid keys from input_dict
            correct_type = type(default_value)
            value = input_dict[key]
            if not _has_valid_type(value, correct_type):
                logging.warning(
                    f"Key '{key}' invalid: expected {correct_type}, got {type(value)} '{value}'"
                )
            try:
                output_dict[key] = correct_type(value)  # type: ignore
            except ValueError:
                logging.warning(
                    f"Failed to coerce Key '{key}' with type {type(value)} to correct type {correct_type}"
                )
        else:
            # warn if a key is missing
            logging.warning(f"Key '{key}' missing, using default '{default_value}'")
    for key in input_dict:
        if key not in output_dict:
            # warn if there are any unknown keys in the input dict
            logging.warning(f"Ignoring unknown key '{key}'")
    return output_dict



from __future__ import annotations

import vstt


def default_display_options() -> vstt.vtypes.DisplayOptions:
    return {
        "to_target_paths": True,
        "to_center_paths": True,
        "targets": True,
        "central_target": True,
        "to_target_reaction_time": True,
        "to_center_reaction_time": False,
        "to_target_movement_time": True,
        "to_center_movement_time": False,
        "to_target_time": True,
        "to_center_time": False,
        "to_target_distance": True,
        "to_center_distance": False,
        "to_target_rmse": True,
        "to_center_rmse": False,
        "to_target_success": False,
        "to_center_success": False,
        "area": False,
        "normalized_area": False,
        "peak_velocity": False,
        "peak_acceleration": False,
        "to_target_spatial_error": False,
        "to_center_spatial_error": False,
        "movement_time_at_peak_velocity": False,
        "total_time_at_peak_velocity": False,
        "movement_distance_at_peak_velocity": False,
        "rmse_movement_at_peak_velocity": False,
        "averages": True,
    }


def display_options_labels() -> dict[str, str]:
    return {
        "to_target_paths": "Display cursor paths to target",
        "to_center_paths": "Display cursor paths back to center",
        "targets": "Display targets",
        "central_target": "Display central target",
        "to_target_reaction_time": "Statistic: reaction time to target",
        "to_center_reaction_time": "Statistic: reaction time to center",
        "to_target_movement_time": "Statistic: movement time to target",
        "to_center_movement_time": "Statistic: movement time to center",
        "to_target_time": "Statistic: total time to target",
        "to_center_time": "Statistic: total time to center",
        "to_target_distance": "Statistic: movement distance to target",
        "to_center_distance": "Statistic: movement distance to center",
        "to_target_rmse": "Statistic: RMSE movement to target",
        "to_center_rmse": "Statistic: RMSE movement to center",
        "to_target_success": "Statistic: successful movement to target",
        "to_center_success": "Statistic: successful movement to center",
        "area": "Statistic: the area formed by the paths connecting the target and the center",
        "normalized_area": "Statistic: (the area formed by paths) / (length of the paths)²",
        "peak_velocity": "Statistic: maximum velocity during cursor movement",
        "peak_acceleration": "Statistic: maximum acceleration during cursor movement",
        "to_target_spatial_error": "Statistic: distance from the movement end point to the target",
        "to_center_spatial_error": "Statistic: distance from the movement end point to the center",
        "movement_time_at_peak_velocity": "Statistic: movement time to the point at the peak velocity",
        "total_time_at_peak_velocity": "Statistic: total time to the point at the peak velocity",
        "movement_distance_at_peak_velocity": "Statistic: movement distance to the point at the peak velocity",
        "rmse_movement_at_peak_velocity": "Statistic: RMSE movement to the point at the peak velocity",
        "averages": "Also show statistics averaged over all targets",
    }


def import_display_options(display_options_dict: dict) -> vstt.vtypes.DisplayOptions:
    return vstt.common.import_typed_dict(
        display_options_dict, default_display_options()
    )



from __future__ import annotations

import qt_test_utils as qtu
from psychopy.visual.window import Window

import vstt
from vstt.experiment import Experiment
from vstt.trials_widget import TrialsWidget


def test_trials_widget_no_results(window: Window) -> None:
    widget = TrialsWidget(parent=None, win=window)
    signal_received = qtu.SignalReceived(widget.experiment_modified)
    experiment = Experiment()
    experiment.trial_list = []
    experiment.has_unsaved_changes = False
    assert not signal_received
    assert experiment.trial_handler_with_results is None
    # set an experiment with no trials
    widget.experiment = experiment
    assert widget.experiment is experiment
    assert len(widget.experiment.trial_list) == 0
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is False
    assert widget._btn_move_up.isEnabled() is False
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is False
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    # set an experiment with a single trial & no results
    experiment.trial_list = [vstt.trial.default_trial()]
    assert experiment.trial_handler_with_results is None
    assert experiment.has_unsaved_changes is False
    widget.experiment = experiment
    assert widget.experiment is experiment
    # select first row
    qtu.press_up_key(widget._widget_list_trials)
    assert widget.experiment.trial_list == [vstt.trial.default_trial()]
    assert widget._widget_list_trials.count() == 1
    assert widget._widget_list_trials.currentRow() == 0
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is False
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is True
    assert experiment.trial_handler_with_results is None
    assert experiment.has_unsaved_changes is False
    assert not signal_received
    # set an experiment with three trials & no results
    trials = [
        vstt.trial.default_trial(),
        vstt.trial.default_trial(),
        vstt.trial.default_trial(),
    ]
    trials[0]["weight"] = 0
    trials[1]["weight"] = 1
    trials[2]["weight"] = 2
    experiment.trial_list = trials
    assert experiment.trial_handler_with_results is None
    assert experiment.has_unsaved_changes is False
    assert not signal_received
    widget.experiment = experiment
    assert widget.experiment is experiment
    # select first row
    for _ in range(3):
        qtu.press_up_key(widget._widget_list_trials)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 1, 2]
    assert widget._widget_list_trials.count() == 3
    assert widget._widget_list_trials.currentRow() == 0
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is False
    assert widget._btn_move_down.isEnabled() is True
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    # select middle row
    qtu.press_down_key(widget._widget_list_trials)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 1, 2]
    assert widget._widget_list_trials.count() == 3
    assert widget._widget_list_trials.currentRow() == 1
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is True
    assert widget._btn_move_down.isEnabled() is True
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    # select last row
    qtu.press_down_key(widget._widget_list_trials)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 1, 2]
    assert widget._widget_list_trials.count() == 3
    assert widget._widget_list_trials.currentRow() == 2
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is True
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is False
    assert not signal_received
    # move trial up
    signal_received.clear()
    qtu.click(widget._btn_move_up)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 2, 1]
    assert widget._widget_list_trials.count() == 3
    assert widget._widget_list_trials.currentRow() == 1
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is True
    assert widget._btn_move_down.isEnabled() is True
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is True
    assert signal_received
    # move trial down again
    signal_received.clear()
    qtu.click(widget._btn_move_down)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 1, 2]
    assert widget._widget_list_trials.count() == 3
    assert widget._widget_list_trials.currentRow() == 2
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is True
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is True
    assert signal_received
    # remove last trial
    signal_received.clear()
    qtu.click(widget._btn_remove)
    qtu.press_down_key(widget._widget_list_trials)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 1]
    assert widget._widget_list_trials.count() == 2
    assert widget._widget_list_trials.currentRow() == 1
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is True
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is True
    assert signal_received
    # select first trial
    signal_received.clear()
    qtu.press_up_key(widget._widget_list_trials)
    assert [t["weight"] for t in widget.experiment.trial_list] == [0, 1]
    assert widget._widget_list_trials.count() == 2
    assert widget._widget_list_trials.currentRow() == 0
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is False
    assert widget._btn_move_down.isEnabled() is True
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is True
    assert not signal_received
    # remove first trial
    signal_received.clear()
    qtu.click(widget._btn_remove)
    qtu.press_down_key(widget._widget_list_trials)
    assert [t["weight"] for t in widget.experiment.trial_list] == [1]
    assert widget._widget_list_trials.count() == 1
    assert widget._widget_list_trials.currentRow() == 0
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is True
    assert widget._btn_move_up.isEnabled() is False
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is True
    assert widget.experiment.has_unsaved_changes is True
    assert signal_received
    # remove last trial
    signal_received.clear()
    qtu.click(widget._btn_remove)
    qtu.press_down_key(widget._widget_list_trials)
    assert widget.experiment.trial_list == []
    assert widget._widget_list_trials.count() == 0
    assert widget._btn_add.isEnabled() is True
    assert widget._btn_edit.isEnabled() is False
    assert widget._btn_move_up.isEnabled() is False
    assert widget._btn_move_down.isEnabled() is False
    assert widget._btn_remove.isEnabled() is False
    assert widget.experiment.has_unsaved_changes is True
    assert signal_received


def test_trials_widget_with_results(
    experiment_with_results: Experiment, window: Window
) -> None:
    widget = TrialsWidget(parent=None, win=window)
    signal_received = qtu.SignalReceived(widget.experiment_modified)
    assert not signal_received
    experiment_with_results.has_unsaved_changes = False
    assert experiment_with_results.trial_handler_with_results is not None
    # set an experiment with results
    widget.experiment = experiment_with_results
    # select first row
    for _ in range(3):
        qtu.press_up_key(widget._widget_list_trials)
    # click no when asked if we want to clear the results
    mwt = qtu.ModalWidgetTimer(["N"])
    mwt.start()
    # click on remove button
    qtu.click(widget._btn_remove)
    assert mwt.widget_type == "QMessageBox"
    assert (
        mwt.widget_text
        == "Modifying the trial conditions will clear the existing results. Continue?"
    )
    assert not signal_received
    assert experiment_with_results.has_unsaved_changes is False
    assert experiment_with_results.trial_handler_with_results is not None
    # this time click yes when asked if we want to clear the results
    mwt = qtu.ModalWidgetTimer(["Y"])
    mwt.start()
    # click on remove button
    qtu.click(widget._btn_remove)
    assert mwt.widget_type == "QMessageBox"
    assert (
        mwt.widget_text
        == "Modifying the trial conditions will clear the existing results. Continue?"
    )
    assert signal_received
    assert experiment_with_results.has_unsaved_changes is True
    assert experiment_with_results.trial_handler_with_results is None



from __future__ import annotations

import logging
from typing import Any
from typing import Iterable

import numpy as np
from psychopy.clock import Clock
from psychopy.data import TrialHandlerExt
from psychopy.event import Mouse
from psychopy.hardware.keyboard import Keyboard
from psychopy.sound import Sound
from psychopy.visual.basevisual import BaseVisualStim
from psychopy.visual.elementarray import ElementArrayStim
from psychopy.visual.shape import ShapeStim
from psychopy.visual.window import Window

import vstt.vtypes
from vstt import joystick_wrapper
from vstt import vis
from vstt.experiment import Experiment
from vstt.geom import JoystickPointUpdater
from vstt.geom import PointRotator
from vstt.geom import to_target_dists
from vstt.stats import stats_dataframe


def _get_target_indices(outer_target_index: int, trial: dict[str, Any]) -> list[int]:
    # always include outer target index
    indices = [outer_target_index]
    if trial["add_central_target"] and not trial["automove_cursor_to_center"]:
        # add central target index
        indices.append(trial["num_targets"])
    return indices


class TrialData:
    """Stores the data from the trial that will be stored in the psychopy trial_handler"""

    def __init__(self, trial: dict[str, Any], rng: np.random.Generator):
        self.target_indices = np.fromstring(
            trial["target_indices"], dtype="int", sep=" "
        )
        if trial["target_order"] == "random":
            rng.shuffle(self.target_indices)
        self.target_pos: list[np.ndarray] = []
        self.to_target_timestamps: list[np.ndarray] = []
        self.to_target_num_timestamps_before_visible: list[int] = []
        self.to_center_timestamps: list[np.ndarray] = []
        self.to_center_num_timestamps_before_visible: list[int] = []
        self.to_target_mouse_positions: list[np.ndarray] = []
        self.to_center_mouse_positions: list[np.ndarray] = []
        self.to_target_success: list[bool] = []
        self.to_center_success: list[bool] = []


class TrialManager:
    """Stores the drawable elements and other objects needed during a trial"""

    def __init__(self, win: Window, trial: vstt.vtypes.Trial):
        self.targets = vis.make_targets(
            win,
            trial["num_targets"],
            trial["target_distance"],
            trial["target_size"],
            trial["add_central_target"],
            trial["central_target_size"],
        )
        self.drawables: list[BaseVisualStim | ElementArrayStim] = [self.targets]
        self.target_labels = None
        if trial["show_target_labels"]:
            self.target_labels = vis.make_target_labels(
                win,
                trial["num_targets"],
                trial["target_distance"],
                trial["target_size"],
                trial["target_labels"],
            )
            self.drawables.extend(self.target_labels)
        self.cursor = vis.make_cursor(win, trial["cursor_size"])
        self.cursor.setPos(np.array([0.0, 0.0]))
        if trial["show_cursor"]:
            self.drawables.append(self.cursor)
        self.point_rotator = PointRotator(trial["cursor_rotation_degrees"])
        self.joystick_point_updater = JoystickPointUpdater(
            trial["cursor_rotation_degrees"], trial["joystick_max_speed"], win.size
        )
        self._cursor_path = ShapeStim(
            win, vertices=[(0.0, 0.0)], lineColor="white", closeShape=False
        )
        self._cursor_path_vertices: list[tuple[float, float]] = []
        self.clock = Clock()
        if trial["show_cursor_path"]:
            self.drawables.append(self._cursor_path)
        self.first_target_of_condition_shown = False
        self.most_recent_target_display_time = 0.0
        self.final_target_display_time_previous_trial = 0.0
        self.green_target_index: int | None = None

    def cursor_path_add_vertex(
        self, vertex: tuple[float, float], clear_existing: bool = False
    ) -> None:
        if clear_existing:
            self._cursor_path_vertices = []
        self._cursor_path_vertices.append(vertex)
        self._cursor_path.vertices = self._cursor_path_vertices


def _contains_mixed_length_numpy_arrays(items: Iterable) -> bool:
    return (
        len(set([item.shape if isinstance(item, np.ndarray) else 1 for item in items]))
        > 1
    )


def add_trial_data_to_trial_handler(
    trial_data: TrialData, trial_handler: TrialHandlerExt
) -> None:
    for name, value in vars(trial_data).items():
        dtype = object if _contains_mixed_length_numpy_arrays(value) else None
        trial_handler.addData(name, np.array(value, dtype=dtype))


class MotorTask:
    def __init__(self, experiment: Experiment, win: Window | None = None):
        self.close_window_when_done = False
        self.experiment = experiment
        if win is None:
            win = Window(fullscr=True, units="height")
            self.close_window_when_done = True
        self.win = win
        if not experiment.trial_list:
            return
        self.trial_managers = {
            condition_index: TrialManager(self.win, trial)
            for condition_index, trial in enumerate(experiment.trial_list)
        }
        self.trial_handler = experiment.create_trialhandler()
        self.mouse = Mouse(visible=False, win=win)
        self.kb = Keyboard()
        self.js = joystick_wrapper.get_joystick()
        self.rng = np.random.default_rng()

    def run(self) -> bool:
        if not self.experiment.trial_list:
            return self._clean_up_and_return(False)
        try:
            self._do_trials()
            self.experiment.trial_handler_with_results = self.trial_handler
            self.experiment.stats = stats_dataframe(self.trial_handler)
            self.experiment.has_unsaved_changes = True
            return self._clean_up_and_return(True)
        except vis.MotorTaskCancelledByUser:
            return self._clean_up_and_return(False)
        except Exception as e:
            logging.warning(f"Run task failed with exception {e}")
            logging.exception(e)
            # clean up window before re-raising the exception
            if self.win is not None and self.close_window_when_done:
                self.win.close()
            raise

    def _do_trials(self) -> None:
        self._do_splash_screen()
        condition_trial_indices: list[list[int]] = [
            [] for _ in self.trial_handler.trialList
        ]
        current_condition_index = -1
        current_condition_first_trial_index = 0
        current_condition_clock = Clock()
        current_condition_max_time = 0.0
        self.mouse.setPos((0.0, 0.0))
        current_cursor_pos = (0.0, 0.0)
        for trial in self.trial_handler:
            if self.trial_handler.thisIndex != current_condition_index:
                # starting a new set of conditions
                current_condition_clock.reset()
                current_condition_max_time = trial["condition_timeout"]
                current_condition_index = self.trial_handler.thisIndex
                current_condition_first_trial_index = self.trial_handler.thisTrialN
                trial_manager = self.trial_managers[current_condition_index]
            condition_trial_indices[self.trial_handler.thisIndex].append(
                self.trial_handler.thisTrialN
            )
            if (
                current_condition_max_time <= 0.0
                or current_condition_clock.getTime() < current_condition_max_time
            ):
                # only do the trial if there is still time left for this condition
                current_cursor_pos = self._do_trial(
                    trial,
                    trial_manager,
                    current_cursor_pos,
                    current_condition_max_time - current_condition_clock.getTime(),
                )
            is_final_trial_of_block = (
                len(condition_trial_indices[self.trial_handler.thisIndex])
                == trial["weight"]
            )
            if is_final_trial_of_block and trial["post_block_delay"] > 0:
                vis.display_results(
                    trial["post_block_delay"],
                    trial["enter_to_skip_delay"],
                    trial["show_delay_countdown"],
                    self.trial_handler if trial["post_block_display_results"] else None,
                    self.experiment.display_options,
                    current_condition_first_trial_index,
                    True,
                    self.win,
                    self.mouse,
                    current_cursor_pos,
                )
        if self.win.nDroppedFrames > 0:
            logging.warning(f"Dropped {self.win.nDroppedFrames} frames")

    def _do_splash_screen(self) -> None:
        vis.splash_screen(
            display_time_seconds=self.experiment.metadata["display_duration"],
            enter_to_skip_delay=self.experiment.metadata["enter_to_skip_delay"],
            show_delay_countdown=self.experiment.metadata["show_delay_countdown"],
            metadata=self.experiment.metadata,
            win=self.win,
        )

    def _do_trial(
        self,
        trial: dict[str, Any],
        trial_manager: TrialManager,
        initial_cursor_pos: tuple[float, float],
        condition_timeout: float,
    ) -> tuple[float, float]:
        if trial["use_joystick"] and self.js is None:
            raise RuntimeError("Use joystick option is enabled, but no joystick found.")
        trial_manager.cursor.setPos(initial_cursor_pos)
        trial_manager.final_target_display_time_previous_trial = (
            trial_manager.most_recent_target_display_time
        )
        trial_data = TrialData(trial, self.rng)
        self.win.recordFrameIntervals = True
        trial_manager.clock.reset()
        vis.update_target_colors(
            trial_manager.targets, trial["show_inactive_targets"], None
        )
        if trial["show_target_labels"] and trial_manager.target_labels is not None:
            vis.update_target_label_colors(
                trial_manager.target_labels, trial["show_inactive_targets"], None
            )
        for index in trial_data.target_indices:
            if (
                condition_timeout <= 0.0
                or trial_manager.clock.getTime() < condition_timeout
            ):
                self._do_target(trial, index, trial_manager, trial_data)
        self.win.recordFrameIntervals = False
        if trial["automove_cursor_to_center"]:
            trial_data.to_center_success = [True] * trial["num_targets"]
        if (
            condition_timeout <= 0.0
            or trial_manager.clock.getTime() < condition_timeout
        ):
            # only store trial data if we didn't run out of time for this condition
            add_trial_data_to_trial_handler(trial_data, self.trial_handler)
        if trial["post_trial_delay"] > 0:
            vis.display_results(
                trial["post_trial_delay"],
                trial["enter_to_skip_delay"],
                trial["show_delay_countdown"],
                self.trial_handler if trial["post_trial_display_results"] else None,
                self.experiment.display_options,
                self.trial_handler.thisTrialN,
                False,
                self.win,
                self.mouse,
                trial_manager.cursor.pos,
            )
        return trial_manager.cursor.pos

    def _do_target(
        self, trial: dict[str, Any], index: int, tm: TrialManager, trial_data: TrialData
    ) -> None:
        minimum_window_for_flip = 1.0 / 60.0
        mouse_pos = tm.cursor.pos
        stop_waiting_time = 0.0
        stop_target_time = 0.0
        if trial["fixed_target_intervals"]:
            num_completed_targets = len(trial_data.to_target_timestamps)
            stop_waiting_time = (num_completed_targets + 1) * trial[
                "target_duration"
            ] - tm.final_target_display_time_previous_trial
            stop_target_time = stop_waiting_time + trial["target_duration"]
        for target_index in _get_target_indices(index, trial):
            t0 = tm.clock.getTime()
            is_central_target = target_index == trial["num_targets"]
            mouse_times = []
            mouse_positions = []

            # current target is not yet displayed
            if not is_central_target:
                if trial["automove_cursor_to_center"]:
                    mouse_pos = np.array([0.0, 0.0])
                    self.mouse.setPos(mouse_pos)
                    tm.cursor.setPos(mouse_pos)
                tm.cursor_path_add_vertex(mouse_pos, clear_existing=True)
            if not trial["fixed_target_intervals"]:
                if is_central_target:
                    pre_target_delay = trial["pre_central_target_delay"]
                else:
                    pre_target_delay = trial["pre_target_delay"]
                    if (
                        not tm.first_target_of_condition_shown
                        and trial["pre_first_target_extra_delay"] > 0
                    ):
                        pre_target_delay += trial["pre_first_target_extra_delay"]
                        tm.first_target_of_condition_shown = True
                stop_waiting_time = t0 + pre_target_delay
            if stop_waiting_time >= t0:
                if trial["hide_target_when_reached"]:
                    vis.update_target_colors(
                        tm.targets,
                        trial["show_inactive_targets"],
                        None,
                        tm.green_target_index,
                    )
                    if trial["show_target_labels"] and tm.target_labels is not None:
                        vis.update_target_label_colors(
                            tm.target_labels, trial["show_inactive_targets"], None
                        )
                # ensure we get at least a single flip
                should_continue_waiting = True
                while should_continue_waiting:
                    if trial["freeze_cursor_between_targets"]:
                        self.mouse.setPos(mouse_pos)
                    vis.draw_and_flip(self.win, tm.drawables, self.kb)
                    if not trial["freeze_cursor_between_targets"]:
                        if trial["use_joystick"]:
                            mouse_pos = tm.joystick_point_updater(
                                mouse_pos,
                                (self.js.getX(), self.js.getY()),  # type: ignore
                            )
                        else:
                            mouse_pos = tm.point_rotator(self.mouse.getPos())
                    mouse_times.append(tm.clock.getTime())
                    mouse_positions.append(mouse_pos)
                    if trial["show_cursor"]:
                        tm.cursor.setPos(mouse_pos)
                    if trial["show_cursor_path"]:
                        tm.cursor_path_add_vertex(mouse_pos)
                    should_continue_waiting = (
                        tm.clock.getTime() + minimum_window_for_flip < stop_waiting_time
                    )
            # display current target
            t0 = tm.clock.getTime()
            vis.update_target_colors(
                tm.targets, trial["show_inactive_targets"], target_index
            )
            if trial["show_target_labels"] and tm.target_labels is not None:
                vis.update_target_label_colors(
                    tm.target_labels, trial["show_inactive_targets"], target_index
                )
            if trial["play_sound"]:
                Sound("A", secs=0.160, blockSize=1024, stereo=True).play()
            if is_central_target:
                trial_data.to_center_num_timestamps_before_visible.append(
                    len(mouse_times)
                )
            else:
                trial_data.target_pos.append(tm.targets.xys[target_index])
                trial_data.to_target_num_timestamps_before_visible.append(
                    len(mouse_times)
                )
            target_size = trial["target_size"]
            if is_central_target:
                target_size = trial["central_target_size"]
            if not trial["fixed_target_intervals"]:
                if is_central_target:
                    stop_target_time = t0 + trial["central_target_duration"]
                else:
                    stop_target_time = t0 + trial["target_duration"]
            dist_correct = 1.0
            # ensure we get at least one flip
            should_continue_target = True
            while should_continue_target:
                vis.draw_and_flip(self.win, tm.drawables, self.kb)
                if trial["use_joystick"]:
                    mouse_pos = tm.joystick_point_updater(
                        mouse_pos,
                        (self.js.getX(), self.js.getY()),  # type: ignore
                    )
                else:
                    mouse_pos = tm.point_rotator(self.mouse.getPos())
                if trial["show_cursor"]:
                    tm.cursor.setPos(mouse_pos)
                mouse_times.append(tm.clock.getTime())
                mouse_positions.append(mouse_pos)
                if trial["show_cursor_path"]:
                    tm.cursor_path_add_vertex(mouse_pos)
                dist_correct, dist_any = to_target_dists(
                    mouse_pos,
                    tm.targets.xys,
                    target_index,
                    trial["add_central_target"],
                )
                if trial["ignore_incorrect_targets"] or is_central_target:
                    dist = dist_correct
                else:
                    dist = dist_any
                should_continue_target = (
                    dist > target_size
                    and tm.clock.getTime() + minimum_window_for_flip < stop_target_time
                )
            tm.most_recent_target_display_time = tm.clock.getTime() - stop_waiting_time
            success = (
                dist_correct <= target_size
                and tm.clock.getTime() + minimum_window_for_flip < stop_target_time
            )
            # When the target is reached, turn the color to green
            if success and trial["turn_target_to_green_when_reached"]:
                tm.green_target_index = target_index
            if is_central_target:
                trial_data.to_center_success.append(success)
            else:
                trial_data.to_target_success.append(success)
            if is_central_target:
                trial_data.to_center_timestamps.append(np.array(mouse_times))
                trial_data.to_center_mouse_positions.append(np.array(mouse_positions))
            else:
                trial_data.to_target_timestamps.append(np.array(mouse_times))
                trial_data.to_target_mouse_positions.append(np.array(mouse_positions))

    def _clean_up_and_return(self, return_value: bool) -> bool:
        if self.win is not None and self.close_window_when_done:
            self.win.close()
        return return_value



from __future__ import annotations

import vstt


def default_metadata() -> vstt.vtypes.Metadata:
    return {
        "name": "Experiment",
        "subject": "Joe Smith",
        "date": "2022/01/01",
        "author": "Author",
        "display_title": "Welcome to the experiment",
        "display_text1": "When a target is presented the target circle will turn red.",
        "display_text2": "You will also hear an audible tone",
        "display_text3": "Try to move the cursor to the red circle "
        "as quickly and accurately as you can.",
        "display_text4": "You can abort the experiment at any time "
        "by pressing the Escape key.",
        "display_duration": 60,
        "show_delay_countdown": True,
        "enter_to_skip_delay": True,
    }


def metadata_labels() -> dict:
    return {
        "name": "Experiment Name",
        "subject": "Subject Name",
        "date": "Date",
        "author": "Author",
        "display_title": "Title text to display",
        "display_text1": "Main text line 1",
        "display_text2": "Main text line 2",
        "display_text3": "Main text line 3",
        "display_text4": "Main text line 4",
        "display_duration": "Display duration (seconds)",
        "show_delay_countdown": "Display a countdown",
        "enter_to_skip_delay": "Skip by pressing enter key",
    }


def import_metadata(metadata_dict: dict) -> vstt.vtypes.Metadata:
    return vstt.common.import_typed_dict(metadata_dict, default_metadata())



from __future__ import annotations

import logging

import click
from psychopy.gui.qtgui import ensureQtApp
from qtpy import QtWidgets

from vstt.gui import Gui


@click.command()
@click.argument("filename", required=False)
def main(filename: str | None) -> None:
    logging.basicConfig(
        format="%(levelname)s %(module)s.%(funcName)s.%(lineno)d :: %(message)s"
    )
    ensureQtApp()
    app = QtWidgets.QApplication.instance()
    assert app is not None
    gui = Gui(filename=filename)
    gui.showMaximized()
    app.exec()


if __name__ == "__main__":
    main()



from __future__ import annotations

import logging
import pathlib

import numpy as np
from psychopy.visual.window import Window
from qtpy import QtWidgets
from qtpy.compat import getsavefilename

from vstt.experiment import Experiment
from vstt.vis import display_results


class ResultsWidget(QtWidgets.QWidget):
    def __init__(
        self, parent: QtWidgets.QWidget | None = None, win: Window | None = None
    ):
        super().__init__(parent)
        self._win = win
        self._experiment = Experiment()

        outer_layout = QtWidgets.QVBoxLayout()
        group_box = QtWidgets.QGroupBox("Results")
        outer_layout.addWidget(group_box)
        inner_layout = QtWidgets.QGridLayout()
        group_box.setLayout(inner_layout)
        self._list_trial_row_to_trial_index: list[int] = []
        self._list_trials = QtWidgets.QListWidget()
        self._list_trials.currentRowChanged.connect(self._row_changed)
        self._list_trials.itemDoubleClicked.connect(self._btn_display_trial_clicked)
        inner_layout.addWidget(self._list_trials, 0, 0, 1, 4)
        self._btn_display_trial = QtWidgets.QPushButton("View Trial")
        self._btn_display_trial.clicked.connect(self._btn_display_trial_clicked)
        inner_layout.addWidget(self._btn_display_trial, 1, 0)
        self._btn_screenshot_trial = QtWidgets.QPushButton("Trial Image")
        self._btn_screenshot_trial.clicked.connect(self._btn_screenshot_trial_clicked)
        inner_layout.addWidget(self._btn_screenshot_trial, 1, 1)
        self._btn_display_condition = QtWidgets.QPushButton("View Condition")
        self._btn_display_condition.clicked.connect(self._btn_display_condition_clicked)
        inner_layout.addWidget(self._btn_display_condition, 1, 2)
        self._btn_screenshot_condition = QtWidgets.QPushButton("Condition Image")
        self._btn_screenshot_condition.clicked.connect(
            self._btn_screenshot_condition_clicked
        )
        inner_layout.addWidget(self._btn_screenshot_condition, 1, 3)
        self.setLayout(outer_layout)
        self._row_changed()

    def _is_valid(self, row: int) -> bool:
        if (
            self._list_trials.count() > 0
            and self.experiment.trial_handler_with_results is None
        ):
            # experiment results have been cleared - re-assign experiment to update widget
            self.experiment = self._experiment
            return False
        return (
            self.experiment.trial_handler_with_results is not None
            and 0 <= row < self._list_trials.count()
        )

    def _row_changed(self) -> None:
        valid = self._is_valid(self._list_trials.currentRow())
        self._btn_display_trial.setEnabled(valid)
        self._btn_screenshot_trial.setEnabled(valid)
        self._btn_display_condition.setEnabled(valid)
        self._btn_screenshot_condition.setEnabled(valid)

    def _display_results(
        self, all_trials_for_this_condition: bool, screenshot: bool
    ) -> None:
        row = self._list_trials.currentRow()
        if not self._is_valid(row):
            return
        screenshot_image = display_results(
            60,
            True,
            False,
            self._experiment.trial_handler_with_results,
            self._experiment.display_options,
            self._list_trial_row_to_trial_index[row],
            all_trials_for_this_condition,
            win=self._win,
            return_screenshot=screenshot,
        )
        if screenshot:
            if screenshot_image is None:
                QtWidgets.QMessageBox.critical(
                    self,
                    "Screenshot error",
                    "Could not take screenshot",
                )
                return
            filename, _ = getsavefilename(
                self,
                "Save screenshot",
                "screenshot.png",
                "Png image file (*.png)",
            )
            if filename == "":
                return
            try:
                screenshot_image.save(pathlib.Path(filename).with_suffix(".png"))
            except Exception as e:
                logging.warning(f"Failed to save file {filename}: {e}")
                QtWidgets.QMessageBox.critical(
                    self,
                    "File save error",
                    f"Could not save file '{filename}'",
                )

    def _btn_display_trial_clicked(self) -> None:
        self._display_results(False, False)

    def _btn_screenshot_trial_clicked(self) -> None:
        self._display_results(False, True)

    def _btn_display_condition_clicked(self) -> None:
        self._display_results(True, False)

    def _btn_screenshot_condition_clicked(self) -> None:
        self._display_results(True, True)

    @property
    def experiment(self) -> Experiment:
        return self._experiment

    @experiment.setter
    def experiment(self, experiment: Experiment) -> None:
        self._list_trials.clear()
        self._list_trial_row_to_trial_index.clear()
        self._experiment = experiment
        trial_handler = experiment.trial_handler_with_results
        if trial_handler is None:
            return
        # assume 1 repetition of experiment
        rep_index = 0
        for trial_index, condition_index in enumerate(trial_handler.sequenceIndices):
            # trials that have no data have a default string instead of a numpy array
            if (
                type(trial_handler.data["target_indices"][(trial_index, rep_index)])
                is np.ndarray
            ):
                self._list_trials.addItem(
                    f"Trial {trial_index} [Condition {condition_index[rep_index]}]"
                )
                self._list_trial_row_to_trial_index.append(trial_index)
